////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma once

#ifndef BASE_INTEGRATOR_H
#define BASE_INTEGRATOR_H

#include "math/bit.h"
#include "math/randsampler.h"
#include "math/random.h"
#include "geometry/ray.h"
#include "geometry/light.h"
#include "geometry/filter.h"
#include "graph/color.h"

class CFARM;
class D3Layer;
class VVoxel;
class Geometry;

class CBaseIntegrator
{
public:
	enum { cSpeed, cQuality };

	CBaseIntegrator();
	virtual ~CBaseIntegrator( void );

	// basic initialization
	virtual void Init( CFARM* pRenderer, D3Layer* pVisLayer, VVoxel* pVoxelLayer, Geometry* pGeometry );

	// main integrator routine returning the luminance for the passed ray & rendering pass
	virtual Color Li( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 ) = 0;

	bool					IsInitialized() const { return m_bInitialized; }
	void					SetInitialized( bool bFlag ) { m_bInitialized = bFlag; };

	// std. get/set functions of control attributes
	void          SetNrOfMCSamples( int Val ) { m_NrOfMCSamples = Val; };
	int           GetNrOfMCSamples() const { return m_NrOfMCSamples; };
	void          SetNrOfVolLightMCSamples( int Val ) { m_NrOfVolLightMCSamples = Val; };
	int           GetNrOfVolLightMCSamples() const { return m_NrOfVolLightMCSamples; };
	void          SetNrOfIndirectLightingMCSamples( int Val ) { m_NrOfIndirectLightingSamples = Val; };
	int           GetNrOfIndirectLightingMCSamples() const { return m_NrOfIndirectLightingSamples; };
	int           GetMinSamplesBeforeRR() const { return m_MinSamplesBeforeRussianRoulette; };
	void          SetMinSamplesBeforeRR( int NrOfSamples ) { m_MinSamplesBeforeRussianRoulette = NrOfSamples; };
	float         GetRRSurvivalProbability() const { return m_RussianRouletteSurvival; };
	void          SetRRSurvivalProbability( float Val ) { m_RussianRouletteSurvival = Val; };

	// get/set render optimization mode
	int					  GetRenderMode() const { return m_RenderMode; };
	virtual void	SetRenderMode( int RenderMode, int  QualityLevel ) { m_RenderMode = RenderMode; m_RenderQuality = QualityLevel; }
	int						GetRenderQuality() const { return m_RenderQuality; };

protected:
	virtual bool  RussianRoulette( float SurvivalProbability );
	virtual bool  RussianRoulette() { return RussianRoulette( m_RussianRouletteSurvival ); };

protected:
	CBaseIntegrator( const CBaseIntegrator& rhs );
	CBaseIntegrator& operator=( const CBaseIntegrator& rhs );


protected:
	Random								m_Random;
	CFARM*								m_pRenderer;
	D3Layer*							m_pVisLayer;
	VVoxel*								m_pVoxelLayer;
	const Geometry*				m_pGeometry;

	int										m_NrOfMCSamples;
	int										m_MinSamplesBeforeRussianRoulette;
	float									m_RussianRouletteSurvival;
	int										m_NrOfVolLightMCSamples;
	int										m_NrOfIndirectLightingSamples;

	int										m_RenderMode;						// rendermode = speed or quality. Default = speed
	int										m_RenderQuality;				// differentiated render quality level (0 = basic/speed, >=1 different qualities)

	bool									m_bInitialized;

};


#endif
// vim:ts=2
