////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdfspectransmission.h"
#include "bxdffresnel.h"

#include <algorithm>

CBxDFSpecTransmission::CBxDFSpecTransmission( const Color& TransmissionColor /* = Color( 1.0, 1.0, 1.0 ) ) */ ) : CBxDFBase( static_cast<BxDFType>( BSDF_TRANSMISSION | BSDF_SPECULAR) ),
                                                 m_pFresnel( new CFresnelDiel), m_TransmissionColor( TransmissionColor )
{

}

CBxDFSpecTransmission::~CBxDFSpecTransmission(void)
{
  delete m_pFresnel;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFSpecTransmission::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Power ); UNUSED( LightColor ); UNUSED( Shade );

  if( Type == CBxDFBase::BSDF_TRANSMISSION )
    return Sample_f_Transmission( Out, In, Pdf, Normal, RenderMat, Sampler );

  return 0.0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFSpecTransmission::Sample_f_Transmission( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Sampler );

	Vector RefrNormal( Normal );
	Vector OutDir = Out.dir();

	// Figure out which $\eta$ is incident and which is transmitted
	double cosI = -RefrNormal.dot( OutDir );
	bool   bEntering = cosI > 0.;
	double EtaI = 1.0;			// to be physically precise this IOR should be carried by the outgoing ray --> TODO:
	double EtaT = RenderMat.ior();
	double n = 0.0;

	// check if the ray is incident, otherwise reverse the object normal
	if( !bEntering ) {
		std::swap( EtaI, EtaT );
		EtaT = 1.0;                     // as a first assumption we go from the dense medium into air/vacuum
		RefrNormal = RefrNormal * -1.0;
	}

	n = EtaI / EtaT;
	Pdf = 1.0;

	double cosT2 = 1.0 - n * n * (1.0 - cosI * cosI);
	if( cosT2 > 0.0f ) {
		Vector T = (OutDir * n) + RefrNormal * (n * cosI - sqrt( cosT2 )); // create refraction ray
		// TODO: removed the slow the calculation of the square root
		// Vector T = ( OutDir + RefrNormal * cosI) * n - RefrNormal * cosI;

		In = Out;
		In.init();
		In.incDepth();	// to avoid infinite loops

		RaySegment NewSegment( Out.hit( SMALL3D4 ), T, Out.prevZone() );
		NewSegment.tmin = SMALL3D3;
		NewSegment.tmax = INFINITE;

		In.push( NewSegment );


	}
	else {  // total reflection
		In.depth = 32000;                                  // flag ray to be terminated
		return 0.0;
	}


	// Fresnel transmission,
	double Rho = 1.0 / fabs( cosI ) * (EtaI * EtaI) / (EtaT * EtaT) * (1.0 - m_pFresnel->Evaluate( cosI, EtaI, EtaT ));
	return Rho;

}
// vim:ts=2
