////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef VOXEL_INTEGRATOR_H
#define VOXEL_INTEGRATOR_H

#include "baseintegrator.h"
#include "graph/color.h"
#include "matphong.h"

#include <vector>
#include <queue>
#include <deque>
#include <string>

class D3Layer;

const int cMaxGradientCacheSize = 128;	// max. nr of gradients that we cache when rendering voxels

// STL queues are by default not directly iterable but one can provide his own
// underlying container. For our caching class we combine this.
template<typename T, typename Container = std::deque<T> >
class CacheContainer: public std::queue<T, Container>
{
public:
	typedef typename Container::iterator iterator;
	typedef typename Container::const_iterator const_iterator;

	iterator begin() { return this->c.begin(); }
	iterator end() { return this->c.end(); }
	const_iterator begin() const { return this->c.begin(); }
	const_iterator end() const { return this->c.end(); }
};

class CVoxelIntegrator : public CBaseIntegrator
{
public:
	CVoxelIntegrator();
	virtual ~CVoxelIntegrator( void );

	// initialize options configured in the optional config object in FARM
	void InitFromConfig();

	// set transfer function
	void SetTransferFunction( int Index ) { assert( Index >= 0 && Index < 3 );  m_TransferFuncIndex = Index; }
	int  GetTransferFunction() const { return m_TransferFuncIndex; };

	// enable/disable max. intensity projection rendering mode
	void EnableMIPRender( bool bFlag = true, double MinThreshold = -1e99) { m_bMIPProjection = bFlag; m_MinMIPThreshold = MinThreshold; }

	// main integrator routine returning the luminance for the passed ray & rendering pass
	virtual Color Li( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 );

	// v-ray depth
	unsigned int	GetMaxVRayDepth() const { return m_MaxVRayDepth; };
	void					SetMaxVRayDepth( int RayDepth ) { m_MaxVRayDepth = RayDepth; };

	// derive the transparency from the hit ray
	Color					RayToTransparency( Ray* pRay );

	// retrieve render pass of treated by this integrator
	int						GetRenderPass() const { return m_RenderPass; };

	// set the threshold below which ray traversal is terminated
	void					SetMinTransmission( const Color& Val ) { m_TransmissionThreshold = Val; };
	Color					GetMinTransmission() const { return m_TransmissionThreshold; };

	// enable/disable shadows during shading
	void					EnableShadows( bool bFlag = true ) { m_bShadows = bFlag; };
	bool					AreShadowsEnabled() const { return m_bShadows; };

	// treat optical absorption
	void					SetOpticalDensity( double Val ) { assert( Val >= 0.0 ); m_OpticalDensity = Val; }
	double				GetOpticalDensity() const { return m_OpticalDensity; };

	// treat shininess for shading
	void					SetShineExp( double Val ) { m_Shine = std::max( 0.0, Val ); };
	double				GetShineExp() const { return m_Shine; }

		// configure ambient occlusion
	void SetAOCParameters( double MinIntensity, double MaxIntensity, double MaxDist, int NrOfSamples )
	{
		m_AOCMinIntensity = MinIntensity;
		m_AOCMaxIntensity = MaxIntensity;
		m_AOCMaxDist = MaxDist;
		m_NrOfAOCSamples = NrOfSamples;
	}


	// define render mode & adjust ambient normalization accordingly
	virtual void	SetRenderMode( int RenderMode, int  QualityLevel );


protected:
	// transmission composition rendering
	virtual Color Li_Transmission( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 );

	// maximum projection rendering
	virtual void  Li_MIP( GeometryEngine* pEngine, Ray* pRay, Color& CurVoxelColor, Color& CurVoxelTransparency);

	// determine the gradient vector for the current voxel region - this is cached!
	Vector				CalcNormalizedVoxelGradient( Ray* pRay );

	// calc the interpolated hit normal based on cubic spline interpolation of gradients
	Vector				CubicNormalInterpolation( Ray* pRay );

	// determine (unnormalized!) voxel gradient for the voxel i, j, k -- this is not cached!!
	void					CalcVoxelGradient( int i, int j, int k, Vector& GradientVec );

	// caching of gradients
	void					AddGradientToCache( int RegionNr, const Vector& Gradient );
	bool					GetNormalizedGradientFromCache( int RegionNr, Vector& GradientVec );

	// main driver of the ray marcher, trying to identify the voxel region based on the passed ray's position
	bool					IdentifyVoxelRegion( Ray *pRay );

	// determine the shading normal based on the data gradient
	Vector				GetShadingNormal( Ray* pRay );

	// get color of the current voxel
	void					GetVoxelColor( Ray* pRay, Color& ShadingColor );

	// tri-linear interpolation of a voxel traversed by a ray
	void					TrilinearVoxelInterpolation( Ray* pRay, Color& ShadingColor );

	// cubic spline interpolation of a voxel traversed by a ray
	void					CubicVoxelInterpolation( Ray* pRay, Color& ShadingColor );

	// calc local ambient occlusion factor of an opaque voxel
	double				CalcAmbientOcclusion( GeometryEngine* pEngine, Ray* pRay );

	// march along a passed ray until we find the next opaque voxel and return the distance
	double				MarchToNextOpaqueVoxel( Ray* pRay );

	// march along a passed ray and integrate the transparency until the end of the ray is reached or an opaque object has been found
	// NOTE: The IntegratedTransparency passed in the beginning should be Color(1, 1, 1)!
	void					IntegrateTransparency( Ray* pRay, Color& IntegratedTransparency );

	// helper
	bool					ExportColorHisto( const std::string& FileName );

protected:
	int									m_MaxVRayDepth;		// maximum voxel ray depth
	int									m_RenderPass;

	Color								m_TransmissionThreshold;	// minimum light transmission threshold (gray scale [0,1]) below wich
																								// no further branching is performed

	double							m_OpticalDensity;			// optical absorption factor applied via Beer's law
	double							m_Shine;									// optional shininess for the shading of opaque surfaces
	double							m_RayMarchDelta;					// delta when marching rays through a voxel volume
	double							m_AOCRayMarchDelta;				// delta when marching ambient occlusion rays

	bool								m_bShadows;								// flag to control if shadows are taken into account (false = shadows are ignored)

	double							m_AmbientNormalization;		// normalization for ambient component (is function of quality - see SetRenderMode())

	double							m_AOCMinIntensity;
	double							m_AOCMaxIntensity;
	double							m_AOCMaxDist;
	int									m_NrOfAOCSamples;

	int									m_TransferFuncIndex;			// index of the opacity transfer function we are currently using

	bool								m_bMIPProjection;					// maximum projection rendering on/off
	double							m_MinMIPThreshold;

	CacheContainer<std::pair<int, Vector> > m_GradientCache;
};

#endif
// vim:ts=2:sw=2
