/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "gbody.h"

#include "mesh.h"
#include "arbbody.h"
#include "quabody.h"
#include "boxbody.h"
#include "cylbody.h"
#include "xyzbody.h"
#include "conebody.h"
#include "planebody.h"
#include "torusbody.h"
#include "spherebody.h"

#include "math/bmath.h"
#include "math/conic.h"
#include "math/matrix3.h"

using std::cout;
using std::endl;

GNILBody GBody::tnull;
GUniverseBody GBody::tuniverse;
GOPRBody GBody::tleft    ("(", Type::LEFT);
GOPRBody GBody::tright   (")", Type::RIGHT);
//GOPRBody GBody::tnot   ("#", Type::NOT);
GOPRBody GBody::tunion   ("|", Type::UNION);
GOPRBody GBody::tplus    ("+", Type::PLUS);
GOPRBody GBody::tminus   ("-", Type::MINUS);
double GBody::infinite = 100000.0;

#if _DEBUG>0
int  GBody::verbose = 0;
bool GBody::check   = false;
#endif

const char* GBody::_typeStr[] = {
	"YZP", "XZP", "XYP", "PLA",
	"PX",  "PY",  "PZ",
	"P3",  "P",

	"RPP", "BOX", "WED",

	"SPH", "S",   "SO",
	"SX",  "SY",  "SZ",
	"ELL",

	"RCC", "REC", "TRC",

	"XCC", "YCC", "ZCC",
	"C/X", "C/Y", "C/Z",
	"CX",  "CY",  "CZ",

	"K/X", "K/Y", "K/Z",
	"KX", " KY",  "KZ",

	"XEC", "YEC", "ZEC",

	"ARB",
	"RHP",
	"QUA",
	"GQ",
	"SQ",

	"X", "Y", "Z",

	"TRX", "TRY", "TRZ",
	"TX",  "TY",  "TZ",

	"*",		// nil space
	"@",		// universe
	"err",		// error body
	"(", ")",	// parenthesis
	"#",		// not=complement operator
	"|",		// logical union/or
	"+", "-",	// intersection / subtraction
};

/*
 * locationWrt matrix
 *    X = completed
 *    - = to be done (if possible)
 *    ? = tried but abandoned
 *    b = through bbox preferable
 * row is the first item checked against the column. Always the smallest body index
 * check against the bigger ones
 *
 *         bbox plane  box WED   sphere ELL  RCC REC TRC   cyl ellcyl   ARB RHP quad torus  +/-
 *  bbox     X    X     -   -      -    -     -   -   -     -    -       -   -    -    -     -
 *  plane         X     X   X      X    -     -   -   -     X    -       X   -    -    -     X
 *  box                 b   b      ?    -     -   -   -     ?    -       b   -    -    -     -
 *  WED                            -    -     -   -   -     -    -       -   -    -    -     -
 *  sphere                         X    -     -   -   -     X    X       -   -    -    -     -
 *  ELL                                 -     -   -   -     -    -       -   -    -    -     -
 *  RCC                                       -   -   -     -    -       -   -    -    -     -
 *  REC                                           -   -     -    -       -   -    -    -     -
 *  TRC                                               -     -    -       -   -    -    -     -
 *  cylinder                                                X    -       -   -    -    -     -
 *  ellcyl                                                       -       -   -    -    -     -
 *  ARB                                                                  b   -    -    -     -
 *  RHP                                                                      -    -    -     -
 *  quad                                                                          -    -     -
 *  torus                                                                              -     -
 *  intersection                                                                             -
 *
 * Reference for various intersections: @see http://www.realtimerendering.com/intersections.html
 */

static int zoneCompare(GZone* const& a, GZone* const& b)
{
	if ((size_t)a < (size_t)b)
		return -1;
	else
	if ((size_t)a > (size_t)b)
		return  1;
	else
		return  0;
} // zoneCompare

/* ============================== GBody ================================ */
/** Geometrical Body */
GBody::GBody(const std::string& aname, const Type atype) :
	_name(aname),
	_id(-1),
	_type(atype),
	_color(0xFF00FF),
	_width(0),
	_generation(0),
	_nQ(0),
	_hasMatrix(false),
	show(0),
	_mesh(nullptr),
	X(Vector::Xo),
	Y(Vector::Yo),
	Z(Vector::Zo),
	xlen(0.0),
	ylen(0.0),
	zlen(0.0),
	_bboxState(BBox::State::Invalid)
	//_obboxState(BBox::State::Invalid)
{
	_matrix.identity();
	_invMatrix.identity();
	zones.compare(zoneCompare);
	save();
} // GBody

/** GBody */
GBody::~GBody()
{
	if (_mesh) delete _mesh;
} // GBody

/** @return memory used by body */
size_t GBody::memory() const
{
	return  sizeof(*this)
	      + ((_mesh!=nullptr)?_mesh->memory():0)
	      + zones.memory();
} // memory

/** save */
void GBody::save()
{
	stype = _type;	// save variables
	SP  = P;
	SPo = Po;
	SX  = X;
	SY  = Y;
	SZ  = Z;
	sxlen = xlen;
	sylen = ylen;
	szlen = zlen;
	sshow = show;
} // save

/** restore */
void GBody::restore()
{
	resetBBox();
	P  = SP;
	Po = SPo;
	X  = SX;
	Y  = SY;
	Z  = SZ;
	xlen = sxlen;
	ylen = sylen;
	zlen = szlen;
	show = sshow;
	_type = stype;
} // restore

/** set position
 * @param x,y,z	new position to set
 */
void GBody::position(const Point& r)
{
	resetBBox();
	P  = _hasMatrix? _invMatrix*r : r;
	Po = P + zlen*Z;
} // position

/** @return transformed position */
Point GBody::position() const
{
	return _hasMatrix? _matrix*P : P;
} // position

/** @return transformed X vector */
Vector GBody::vectorX() const
{
	return _hasMatrix? _matrix*X : X;
} // vectorX

/** @return transformed Y vector */
Vector GBody::vectorY() const
{
	return _hasMatrix? _matrix*Y : Y;
} // vectorY

/** @return transformed Z vector */
Vector GBody::vectorZ() const
{
	return _hasMatrix? _matrix*Z : Z;
} // vectorZ

/** @return transformed position */
Point GBody::node(const int n) const
{
	Point r;
	switch (n) {
		case 0: r = P;
			break;
		case 1: r = P+zlen*Z;
			break;
		case 2: r = P+xlen*X;
			break;
		case 3: r = P+ylen*Y;
			break;
	}
	return _hasMatrix? _matrix*r : r;
} // node

/** @return saved position of body */
Point GBody::savedPosition() const
{
	return _hasMatrix? _matrix*SP : SP;
} // savedPosition

/**  @return closest point, edge, face
 * Values:
 *	0	position of body
 *	1-6	closest quad of body
 *	10-19	closest handler of body
 */
int GBody::closest(const Point& r, const double d, const Vector& w) const
{
	double d2 = d*d;

	Point T = _hasMatrix? _invMatrix*r : r;
	if ((P-T).length2() <= d2) return NodeId_0;

	// Check possible nodes
	if (nodes()>1 and zlen>0.0 and (P+zlen*Z-T).length2() <= d2) return NodeId_Z;
	if (nodes()>2 and xlen>0.0 and (P+xlen*X-T).length2() <= d2) return NodeId_X;
	if (nodes()>3 and ylen>0.0 and (P+ylen*Y-T).length2() <= d2) return NodeId_Y;

	int	close = -1;
	double	dmin = INFINITE;
	for (int i=0; i<nQ(); i++) {
		double dq = Q(i)(r);
		Vector g = Q(i).grad(r);
		double gg = g.normalize();
		dq = Abs(dq/gg);
		// Check if smaller but also normal is not aligned with the w axis
		if (dq<dmin and !Eq0(Abs(g*w)-1.0,0.000001)) {
			close = i;
			dmin = dq;
		}
	}

	return close+1;
} // closest

/** round to zero all parameters with small values */
void GBody::round0(double* what, int n, const double eps) const
{
	for (int i=0; i<n; i++)
		if (Eq0(what[i],eps))
			what[i] = 0.0;
} // round0

/** matrix - set transformation matrix */
void GBody::matrix(const Matrix4& M)
{
	resetBBox();
	_hasMatrix = true;
	_matrix.copy(M);
	_matrix.fix();
	_invMatrix.inverse(M);		// Invert matrix
	_invMatrix.fix();
} // matrix

/** transform - transform quadrics with matrix */
void GBody::transform()
{
	for (int i=0; i<_nQ; i++) {
#if defined(_DUMP) && defined(STRONGDEBUG)
		if (show) {
			cout << "*-* GBody::transform " << name() << ":" << i << endl;
			cout << "*-*\t Q[" << i << "] = " << Q(i) << endl;
			cout << "*-*\t M=" << endl << _invMatrix << endl;
		}
#endif
		_Q[i].transform(_invMatrix, _matrix);
		_Q[i].normalize();
		DUMP(if (show) cout << "*-*\ttQ["<< i << "] = " << Q(i) << endl);
	}

	// FIXME
	// for planes and infinite bodies we should have a special transform
	// or ask to re-create the mesh!
	if (_mesh) _mesh->transform(_matrix);
} // transform

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GBody::move(int item, const Point& r, const Vector&)
{
	double len, angle;
	const double EPS = 1E-7;
	resetBBox();
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= SP;

	switch (item) {
		case NodeId_Z:	// Move Z freely
			len = T.normalize();
			if (Eq0(len,EPS)) return;
			angle = acos(Range(-1.0, SZ*T, 1.0));
			if (!Eq0(angle,EPS) and !Eq(angle,PI,EPS)) {
				Vector axis = SZ ^ T;
				axis.normalize();
				Matrix4	rot;
				rot.rotate(angle, axis);
				//rot.fix();
				X = rot * SX;
				Y = rot * SY;
				Z = rot * SZ;
			} else {
				X = SX;
				Y = SY;
				Z = angle<PIover2? SZ : -SZ;
			}
			zlen = len;
			Po = P + zlen*Z;
			break;

		case NodeId_X:	// Move X freely
			len = T.normalize();
			if (Eq0(len,EPS)) return;
			angle = acos(Range(-1.0, SX*T, 1.0));
			if (!Eq0(angle,EPS) and !Eq(angle,PI,EPS)) {
				Vector axis = SX ^ T;
				axis.normalize();
				Matrix4	rot;
				rot.rotate(angle, axis);
				//rot.fix();
				X = rot * SX;
				Y = rot * SY;
				Z = rot * SZ;
			} else {
				X = angle<PIover2? SX : -SX;
				Y = SY;
				Z = SZ;
			}
			xlen = len;
			Po   = P + zlen*Z;
			break;

		case NodeId_Y:	// Move Y freely
			len = T.normalize();
			if (Eq0(len,EPS)) return;
			angle = acos(Range(-1.0, SY*T, 1.0));
			if (!Eq0(angle,EPS) and !Eq(angle,PI,EPS)) {
				Vector axis = SY ^ T;
				axis.normalize();
				Matrix4	rot;
				rot.rotate(angle, axis);
				//rot.fix();
				X = rot * SX;
				Y = rot * SY;
				Z = rot * SZ;
			} else {
				X = SX;
				Y = angle<PIover2? SY : -SY;
				Z = SZ;
			}
			ylen = len;
			Po   = P + zlen*Z;
			break;

		case NodeId_SZ:	// Move Z along SZ axis
			zlen = T * SZ;
			if (zlen<0.0) {
				zlen = -zlen;
				Z = -SZ;
			} else
				Z = SZ;
			Po = P + zlen*Z;
			break;

		case NodeId_SX:	// Move X along SX axis
			xlen = T * SX;
			if (xlen<0.0) {
				xlen = -xlen;
				X = -SX;
			} else
				X = SX;
			Po = P + zlen*Z;
			break;

		case NodeId_SY:	// Move Y along SY axis
			ylen = T * SY;
			if (ylen<0.0) {
				ylen = -ylen;
				Y = -SY;
			} else
				Y = SY;
			Po = P + zlen*Z;
			break;

		default:
			position(r);
	}
	roundVector(P);
	roundVector(Po);
	roundVector(X);
	roundVector(Y);
	roundVector(Z);
	roundFloat(xlen);
	roundFloat(ylen);
	roundFloat(zlen);
} // move

/** rotate body around an axis and current body position
 * @param angle	to rotate
 * @param axis	to rotate
 */
void GBody::rotate(const double angle, const Vector& axis)
{
	Matrix4	rot;
	rot.rotate(angle,axis);

	X = rot * X;
	Y = rot * Y;
	Z = rot * Z;

	Po = P + zlen*Z;

	roundVector(Po);
	roundVector(X);
	roundVector(Y);
	roundVector(Z);
	resetBBox();
} // rotate

/** bbox */
void GBody::bbox(const BBox& bb)
{
	_bboxState = BBox::State::UserDefined;
	_bbox = bb;
} // bbox

/** createMesh */
void GBody::createMesh()
{
	if (_mesh==nullptr)
		_mesh = new Mesh;
} // createMesh

/**
 * @param body		other body to check location against
 * @return location of this body with respect to another body
 *
 * WARNING: Internally it calls the _locationWrt of the body
 * with lower-index type() to the one with higher
 */
GBody::Location GBody::locationWrt(const GBody* body) const
{
	if (type() > body->type()) {
		// Invert the location
		Location loc = body->_locationWrt(this);
		if (loc == Location::AinB)
			return Location::BinA;
		else
		if (loc == Location::BinA)
			return Location::AinB;
		return loc;
	} else
		return _locationWrt(body);
} // locationWrt

/** check location of this bounding box against the other body's bounding box
 * @param body	other body to check location against
 * @return location of this body bbox with respect to another body bbox
 */
GBody::Location GBody::bbLocationWrt(const GBody* body) const
{
	return GBody::_locationWrt(body);
} // bbLocationWrt

/**
 * @param body	other body to check location against
 * @return location of this body with respect to another body
 */
GBody::Location GBody::_locationWrt(const GBody* body) const
{
	const BBox& Abb = bbox();
	const BBox& Bbb = body->bbox();
	if (Abb.overlap(Bbb))
		return Location::OVERLAP;
	return Location::OUTSIDE;
} // _locationWrt

/**
 * Add a quad: Cxx*x^2 + Cyy*y^2 + Czz*z^2 +
 *             Cxy*x*y + Cxz*x*z + Cyz*y*z +
 *              Cx*x   +  Cy*y   +  Cz*z   + C = 0
 *
 *          0       1      2      3
 *     0 / Cxx     Cxy/2  Cxz/2  Cx/2 \
 *     1 | Cxy/2   Cyy    Cyz/2  Cy/2 |
 *     2 | Cxz/2   Cyz/2  Czz    Cz/2 |
 *     3 \ Cx/2    Cy/2   Cz/2   C    /
 */
void GBody::addQuad(const Quad& q)
{
	assert(_nQ<BODYQUADS);
	_Q[_nQ++] = q;
} // addQuad

void GBody::addQuad(double Cxx, double Cyy, double Czz,
		    double Cxy, double Cxz, double Cyz,
		    double Cx,  double Cy,  double Cz,
		    double CC)
{
	assert(_nQ<BODYQUADS);
	_Q[_nQ] = Quad(Cxx,Cyy,Czz,  Cxy,Cxz,Cyz, Cx,Cy,Cz, CC);
	_Q[_nQ].normalize();
	_nQ++;
} // addQuad

void GBody::addQuad(double Cx,  double Cy,  double Cz, double CC)
{
	assert(_nQ<BODYQUADS);
	_Q[_nQ] = Quad(Cx,Cy,Cz, CC);
	_Q[_nQ].normalize();
	_nQ++;
} // addQuad

/** check for orthogonality in X Y Z */
void GBody::checkOrthogonal() const
{
	// check for orthogonality
	if (!Eq0(X*Y,SMALL3*(X.max()+Y.max()+1)))
		throw sformat("Body: %s %s X-Y vectors not perpendicular X.Y=%g",
			cname(), typeStr(), X*Y);
	if (!Eq0(X*Z,SMALL3*(X.max()+Z.max()+1)))
		throw sformat("Body: %s %s X-Z vectors not perpendicular X.Z=%g",
			cname(), typeStr(), X*Z);
	if (!Eq0(Y*Z,SMALL3*(Y.max()+Z.max()+1)))
		throw sformat("Body: %s %s Y-Z vectors not perpendicular Y.Z=%g",
			cname(), typeStr(), Y*Z);
} // checkOrthogonal

/** findXYZ - find an arbitrary XYZ orthogonal system based only
 * on Z information
 */
void GBody::findXYZ()
{
	if (Eq0(Z.length2(),SMALL)) {
		X = Vector::Xo;
		Y = Vector::Yo;
		Z = Vector::Zo;
		throw std::string("Invalid height |Z|=0");
	} else {
		Y = Z.orthogonal();
		// create the rotation matrix
		Y.normalize();
		X = Y ^ Z;
		X.normalize();
	}
} // findXYZ

/** newBody
 * @param aname		name of body
 * @param atype		string type of body
 * @return a new allocated GBody()
 */
GBody* GBody::newBody(const std::string& name, const std::string& type)
{
	//std::string utype = type;
	//std::transform(utype.begin(), utype.end(), type.begin(), std::toupper);
	try {
		     if (type=="XYP")
				return new GPLABody(name, Type::XYP);
		else if (type=="XZP")
				return new GPLABody(name, Type::XZP);
		else if (type=="YZP")
				return new GPLABody(name, Type::YZP);
		else if (type=="PLA")
				return new GPLABody(name, Type::PLA);

		else if (type=="PZ")
				return new GPLABody(name, Type::PZ);
		else if (type=="PY")
				return new GPLABody(name, Type::PY);
		else if (type=="PX")
				return new GPLABody(name, Type::PX);
		else if (type=="P3")
				return new GP3Body(name);
		else if (type=="P")
				return new GPLABody(name, Type::P);

		else if (type=="RPP")
				return new GBOXBody(name, Type::RPP);
		else if (type=="BOX")
				return new GBOXBody(name, Type::BOX);

		else if (type=="SPH")
				return new GSPHBody(name, Type::SPH);
		else if (type=="S")
				return new GSPHBody(name, Type::S);
		else if (type=="SO")
				return new GSPHBody(name, Type::SO);
		else if (type=="SX")
				return new GSPHBody(name, Type::SX);
		else if (type=="SY")
				return new GSPHBody(name, Type::SY);
		else if (type=="SZ")
				return new GSPHBody(name, Type::SZ);

		else if (type=="XCC")
				return new GInfEllCylBody(name, Type::XCC);
		else if (type=="YCC")
				return new GInfEllCylBody(name, Type::YCC);
		else if (type=="ZCC")
				return new GInfEllCylBody(name, Type::ZCC);
		else if (type=="C/X")
				return new GInfEllCylBody(name, Type::C_X);
		else if (type=="C/Y")
				return new GInfEllCylBody(name, Type::C_Y);
		else if (type=="C/Z")
				return new GInfEllCylBody(name, Type::C_Z);

		else if (type=="CX")
				return new GInfEllCylBody(name, Type::CX);
		else if (type=="CY")
				return new GInfEllCylBody(name, Type::CY);
		else if (type=="CZ")
				return new GInfEllCylBody(name, Type::CZ);

		else if (type=="RCC")
				return new GRCCBody(name);
		else if (type=="REC")
				return new GRECBody(name);
		else if (type=="XEC")
				return new GInfEllCylBody(name, Type::XEC);
		else if (type=="YEC")
				return new GInfEllCylBody(name, Type::YEC);
		else if (type=="ZEC")
				return new GInfEllCylBody(name, Type::ZEC);

		else if (type=="K/X")
				return new GConeBody(name, Type::K_X);
		else if (type=="K/Y")
				return new GConeBody(name, Type::K_Y);
		else if (type=="K/Z")
				return new GConeBody(name, Type::K_Z);

		else if (type=="KX")
				return new GConeBody(name, Type::KX);
		else if (type=="KY")
				return new GConeBody(name, Type::KY);
		else if (type=="KZ")
				return new GConeBody(name, Type::KZ);

		else if (type=="TRC")
				return new GTRCBody(name);
		else if (type=="ELL")
				return new GELLBody(name);

		else if (type=="QUA")
				return new GQUABody(name, Type::QUA);
		else if (type=="GQ")
				return new GQUABody(name, Type::GQ);
		else if (type=="SQ")
				return new GQUABody(name, Type::SQ);

		else if (type=="WED" or type=="RAW")
				return new GWEDBody(name);
		else if (type=="ARB")
				return new GARBBody(name);
		//else if (type=="RHP")
		//		return new GRHPBody(name);

		else if (type=="X")
				return new GXYZBody(name, Type::X);
		else if (type=="Y")
				return new GXYZBody(name, Type::Y);
		else if (type=="Z")
				return new GXYZBody(name, Type::Z);

		else if (type=="TRX")
				return new GTorusBody(name, Type::TRX);
		else if (type=="TRY")
				return new GTorusBody(name, Type::TRY);
		else if (type=="TRZ")
				return new GTorusBody(name, Type::TRZ);
		else if (type=="TX")
				return new GTorusBody(name, Type::TX);
		else if (type=="TY")
				return new GTorusBody(name, Type::TY);
		else if (type=="TZ")
				return new GTorusBody(name, Type::TZ);
		else
			return new GERRBody(name,Type::ERR);
	} catch (...) {
		return new GERRBody(name,Type::ERR);
	}
} // newBody

/** newBody
 * @param aname		name of body
 * @param atype		body type
 * @return a new allocated GBody()
 */
GBody* GBody::newBody(const std::string& name, const Type type)
{
	try {
		switch (type) {
			case Type::XYP:
			case Type::XZP:
			case Type::YZP:
			case Type::PLA:
			case Type::PX:
			case Type::PY:
			case Type::PZ:
			case Type::P:   return new GPLABody(name, type);
			case Type::P3:  return new GP3Body(name);

			case Type::RPP: return new GBOXBody(name, type);
			case Type::BOX: return new GBOXBody(name, type);
			case Type::WED: return new GWEDBody(name);

			case Type::SPH:
			case Type::S:
			case Type::SO:
			case Type::SX:
			case Type::SY:
			case Type::SZ:
				return new GSPHBody(name, type);

			case Type::XCC:
			case Type::YCC:
			case Type::ZCC:
			case Type::C_X:
			case Type::C_Y:
			case Type::C_Z:
			case Type::CX:
			case Type::CY:
			case Type::CZ:
			case Type::XEC:
			case Type::YEC:
			case Type::ZEC: return new GInfEllCylBody(name, type);

			case Type::K_X:
			case Type::K_Y:
			case Type::K_Z:
			case Type::KX:
			case Type::KY:
			case Type::KZ:  return new GConeBody(name, type);

			case Type::RCC: return new GRCCBody(name);
			case Type::REC: return new GRECBody(name);
			case Type::TRC: return new GTRCBody(name);

			case Type::ELL: return new GELLBody(name);
			case Type::ARB: return new GARBBody(name);
			//case Type::RHP: return new GRHPBody(name);

			case Type::X:
			case Type::Y:
			case Type::Z:   return new GXYZBody(name, type);

			case Type::QUA:
			case Type::GQ:
			case Type::SQ:  return new GQUABody(name, type);

			case Type::TRX:
			case Type::TRY:
			case Type::TRZ:
			case Type::TX:
			case Type::TY:
			case Type::TZ:
				return new GTorusBody(name,type);

			default:
				return new GERRBody(name,Type::ERR);
		}
	} catch (...) {
		return new GERRBody(name,Type::ERR);
	}
} // newBody

/**
 * Find if a point (x,y,z) is inside the body or not, using the
 * quadratic surface equations and the direction the point is moving
 * @param withinAcc	return any of the quads was checked on the boundary
 * @return: true  inside
 *          false outside
 * XXX I've tried with const Point&p and d but in the end it was slower
 *     than with the plain references x,y,z...
 */
bool GBody::inside(const double x, const double y, const double z,
		   const double u, const double v, const double w,
		   const double* acc,
#ifdef EXPERIMENTAL
		  bool* withinAcc,
#endif
		  const int ignore1, const int ignore2) const
{
#ifdef EXPERIMENTAL
	if (withinAcc) *withinAcc = false;
#endif
	for (int i=nQ(); i--; ) {
		if (i==ignore1 or i==ignore2) continue;
		double q = Q(i).sense(x, y, z);

		// bail out quickly for large values
		if (q > (Abs(x)+Abs(y)+Abs(z)+1)*0.1) return false;

		// The accuracy factor contains also the error from the conversion
		// of the Quadric to Conics
		double a;
		if (acc)
			a = acc[i] * Q(i).acc(x,y,z, SMALL1);
		else
			a = Q(i).acc(x,y,z, SMALL1);

		// FIXME we have to keep the information if the point is ON the surface
		// then to go directly to the gradient check!!!!
		// could result in numerical precision errors for fancy quadrics
#ifdef EXPERIMENTAL
		//if (withinAcc and InRange(-1000.0*a, q, 1000.0*a))
		//	*withinAcc = true;
#endif
		if (q > a) {
			assert(Q(i)(x,y,z) > a);
			return false;
		} else
		if (q >= -a and ignore1<0) {
			assert(Q(i)(x,y,z) >= -a);
			// Do not search direction on the boundary while we are
			// in intersectBody. This might add additional vertices
			// but is better than to ignore vertices
			// When we are on the boundary
			// check dot product with gradient
#ifdef EXPERIMENTAL
			if (withinAcc) *withinAcc = true;
#endif
			double gx, gy, gz;
			Q(i).grad(x, y, z, &gx, &gy, &gz);
			double g = sqrt(gx*gx + gy*gy + gz*gz);
			if (u*gx + v*gy + w*gz > SMALL*g)
				return false;
		}
	}
	return true;
} // inside2D

/**
 * Find if a point (x,y,z) is inside, surface or outside the body, using the
 * quadratic surface equations distance
 * @param x,y,z		point to check
 * @param dsurf		distance tolerance to be considered on surface
 * @return BSide enumerator BInside, BSurface, BOutside
 */
GBody::Side GBody::side(const double x, const double y, const double z,
		   const double dsurf) const
{
	Side iside = Side::In;		// initial guess inside or surface
	for (int i=0; i<nQ(); i++) {
		double d = Q(i).distance(x,y,z);

		// The accuracy factor contains also the error from the conversion
		// of the Quadric to Conics
		double a = Max(Q(i).acc(x, y, z, SMALL1), dsurf);

		if (d > a)	// return immediately outside
			return Side::Out;
		else
		if (d >= -a)	// potentially on surface (unless outside)
			iside = Side::Surface;
	}
	return iside;
} // side

/**
 * Find if a point (x,y,z) is inside the body or not, using the
 * quadratic surface equations.
 * @param x,y,z	location to search
 * @param ignore_a, ignore_b, ignore_c quads to ignore in the computation
 * @return:
 *	true if the point is inside or all the quads were ignored
 *	false if outside
 * WARNING if ignore is not supplied and the point is on the boundary
 *         then it is checked against the normal
 */
bool GBody::inside(const double x, const double y, const double z,
		const Quad* ignore_a, const Quad* ignore_b, const Quad* ignore_c) const
{
	for (int i=0; i<nQ(); i++) {
		const Quad &quad = Q(i);
		if (&quad == ignore_a or
		    &quad == ignore_b or
		    &quad == ignore_c)
			continue;
		double q = quad.sense(x,y,z);

		// Divide by gradient
		double gx, gy, gz, g;
		quad.grad(x, y, z, &gx, &gy, &gz);
		g = sqrt(gx*gx + gy*gy + gz*gz);
		if (g > SMALL) q /= g;
		if (q > quad.acc(x,y,z, SMALL2))
			return false;
	}
	return true;
} // inside

/**
 * Find if a point (x,y,z) is outside the body or not, using the
 * quadratic surface equations.
 * @param x,y,z	location to search
 * @param ignore_a, ignore_b, ignore_c quads to ignore in the computation
 * @return:
 *	true if the point is outside or all the quads were ignored
 *	false if inside
 */
bool GBody::outside(const double x, const double y, const double z,
		    const Quad* ignore_a,
		    const Quad* ignore_b,
		    const Quad* ignore_c) const
{
	for (int i=0; i<nQ(); i++) {
		const Quad &quad = Q(i);
		if (&quad == ignore_a or
		    &quad == ignore_b or
		    &quad == ignore_c)
			continue;
		double q = quad.sense(x,y,z);
		double gx, gy, gz, g;
		quad.grad(x, y, z, &gx, &gy, &gz);
		g = sqrt(gx*gx + gy*gy + gz*gz);
		if (g > SMALL) q /= g;
		if (q < -quad.acc(x,y,z, SMALL2))
			return false;
	}
	return true;
} // outside

/**
 * @param x,y,z	position
 * @return minimum positive distance from all bodies conics
 */
double GBody::distance(const double x, const double y, const double z) const
{
	double dmin = INFINITE;
	for (int i=0; i<nQ(); i++) {
		double q = Abs(Q(i).distance(x,y,z));
		if (q < dmin) dmin = q;
	}
	return dmin;
} // distance

/** calculates the tmin, tmax distance and inverse from the intersection
 * of the ray with the body
 * @param x,y,z		position vector
 * @param u,v,w		direction vector (has to be normalized)
 * @param tmin, tmax	entry and exit point to body
 * @return		true if inverse should be used, false otherwise
 * On intersection tmin<=tmax, otherwise no intersection is found
 *
 * FIXME should return the one/two surfaces that returned the result
 *       to be used for the normal calculation
 */
bool GBody::distance(const double x, const double y, const double z,
		     const double u, const double v, const double w,
		     double* tmin, double* tmax) const
{
	double t2min, t2max;
	bool two = false;
	bool inv = false;

	*tmin = t2min = -INFINITE;
	*tmax = t2max =  INFINITE;

	for (int i=0; i<nQ(); i++) {
		int n;
		double ta, tb;
		if ((n=Q(i).distance(x,y,z, u,v,w, &ta, &tb))) {
			if (n<0) {
				assert(n==-2 and i==0);
				// Only for QUA surface
				if (nQ()==1) {
					// only the QUA are allowed to be concave
					*tmin = ta;
					*tmax = tb;
					return true;
				} else {
					// e.g. TRC
					// two ranges [-inf..ta] and [tb..+inf]
					two = true;
					*tmax = ta;
					t2min = tb;
				}
			} else {
				if (ta > *tmin) *tmin = ta;
				if (tb < *tmax) *tmax = tb;
				if (two) {
					if (ta > t2min) t2min = ta;
					if (tb < t2max) t2max = tb;
					if (t2min>t2max)
						two = false;	// invalid second region
				}
			}
			if (*tmin>*tmax) {	// first region is invalid
				if (two) {
					// move second region to first
					*tmin = t2min;
					*tmax = t2max;
					two  = false;
				}
				if (*tmin>*tmax) {
					// if still first is invalid then return false
					*tmin =  INFINITE;
					*tmax = -INFINITE;
					return false;
				}
			}
		} else
		if (Q(i).sense(x,y,z)>0.0) {
			// no intersection and we are outside
			*tmin =  INFINITE;
			*tmax = -INFINITE;
			return false;
		}
	}
	assert(isTorus() or !two);

	return inv;
} // distance

/** return surface normal at position p(x,y,z)
 * @param x,y,z	position vector
 * @return	normal from the closest surface
 */
Vector GBody::normal(const double x, const double y, const double z) const
{
	double dx, dy, dz;
	int small = -1;
	double qsmall = INFINITE;
	for (int i=0; i<nQ(); i++) {
		double q = Q(i)(x,y,z);
		Q(i).grad(x,y,z, &dx, &dy, &dz);
		double g = sqrt(Sq(dx) + Sq(dy) + Sq(dz));
		if (g > SMALL) {
			g = 1.0 / g;
			q *= g;
		}
		double a = Q(i).acc(x,y,z, SMALL4);
		q = Abs(q);
		if (q<=a)
			return Vector(dx*g, dy*g, dz*g);
		if (q < qsmall) {
			qsmall = q;
			small = i;
		}
	}
	assert(small>=0);

	// normalize and return the smallest value
	Q(small).grad(x,y,z, &dx, &dy, &dz);
	double g = sqrt(Sq(dx) + Sq(dy) + Sq(dz));
	if (g > SMALL) {
		g = 1.0 / g;
		return Vector(dx*g, dy*g, dz*g);
	} else
		return Vector::Zo;
} // normal

/** operator << */
std::ostream& operator << (std::ostream& s, const GBody& body)
{
	double what[30];
	int n = body.get(what);
	s << body.typeStr() << " " << body.name();
	for (int i=0; i<n; i++)
		s << " " << what[i];
	return s;
} /* operator << */
