/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __GEOMETRY_GBODY_H
#define __GEOMETRY_GBODY_H

#include "geo.h"
#include "mesh.h"
#include "math/bbox.h"
#include "math/quad.h"
#include "tool/array.h"
#include "tool/cache.h"
#include "math/obbox.h"
#include "math/matrix4.h"
#include "tool/bstring.h"

class GZone;
class GOPRBody;
class GNILBody;
class GUniverseBody;

// ================================== CBody3D ==================================
// Caching body for 3D operations only
class CBody3D : public CacheItem {
protected:
	bool	tinv;		// if [tmin,tmax] solution should be
				// used in inverse e.g. for concave objects
				// or cones
	double	tmin, tmax;	// ray intersection min, max distances
#ifdef STRONGDEBUG			// STRONG debugging
	Point	p;		// remember position
	Vector	v;		// remember direction
#endif

public:
	CBody3D() : CacheItem() {}
#ifdef STRONGDEBUG
	void	store(	const Point& ap, const Vector& av) {
				p = ap;
				v = av;
			}
	void	store(	const double ax, const double ay, const double az,
			const double au, const double av, const double aw) {
				p.x = ax;
				p.y = ay;
				p.z = az;
				v.x = au;
				v.y = av;
				v.z = aw;
			}
	void	_assert(const Point& ap, const Vector& av) {
				assert(p == ap);
				assert(v == av);
			}
	void	_assert(const double ax, const double ay, const double az,
			const double au, const double av, const double aw) {
				assert(p.x==ax);
				assert(p.y==ay);
				assert(p.z==az);
				assert(v.x==au);
				assert(v.y==av);
				assert(v.z==aw);
			}
#endif
friend class GBody;
friend class GZone;
}; // class CBody3D

#if 0
// ================================= CBody3D2D =================================
// Caching body for 3D and 2D-Viewport operations
class CBody3D2D : public CBody3D {
protected:
	bool	checkInside;	// Result of the last inside check
	int	_checkId;	// instance checked Id
	int*	gCheckId;	// global current check Id
}; // class CBody3D2D
#endif

// =================================== GBody ===================================
/** GBody geometrical body class */
class GBody {
public:
static	const int BODYQUADS  =  6;
static	const int BODYCONICS =  6;
static	const int MAXWHATS   = 30;

	// FIXME merge with Side
	enum class Location : int {
		UNKNOWN = -1,	// Unknown information (FIXME should become zero)
		OUTSIDE =  0,	// The viewport is outside the body
		INSIDE  =  1,	// The viewport is inside the body
		OVERLAP =  2,	// The viewport is overlapping the body
		AinB    =  3,	// A body is inside B	(only from location Wrt)
		BinA    =  4,	// B body is inside A	(only from location Wrt)
	};

	// FIXME merge with Location
	// Possible inside return codes, when no direction is provided
	enum class Side {		// same order as G4 EInside
		Out,
		Surface,
		In
	};

	// WARNING DO NOT change the order the following depend on the order
	// @see _typeStr, __locationWrt, isBody, isAnyBody, isOperator,
	//	isIntersection, isPlane, isInfCylinder, isTorus
	enum class Type {
		YZP, XZP, XYP, PLA,	// planes
		PX,  PY,  PZ,		// equivalent PX, PY, PZ
		P3,  P,			// plane with 3 points and P(a,b,c,d)

		RPP, BOX, WED,		// boxes

		SPH, S,   SO,		// spheres
		SX,  SY,  SZ,
		ELL,			// ellipsoid

		RCC, REC, TRC,		// cylinder

		XCC, YCC, ZCC,		// aligned cylinders
		C_X, C_Y, C_Z,		// equivalent C_X, C_Y, C_Z,
		CX,  CY,  CZ,

		K_X, K_Y, K_Z,		// aligned cones
		KX,  KY,  KZ,

		XEC, YEC, ZEC,		// aligned elliptical cylinders

		ARB,			// arbitrary tetrahedron
		RHP,			// Right hexagonal prism

		QUA,			// generic quadratic
		GQ,			// non parallel to axis
		SQ,			// ellipse, hyper/paraboloid parallel with X,Y,Z axis

		X, Y, Z,		// Generic X,Y,Z that can create almost anything

		TRX, TRY, TRZ,		// tori
		TX,  TY,  TZ,		// synonyms

		/* end of normal bodies */

		NIL,			// * nil space / nothing body=nothing (no ref)
		REGION,			// #region as a body in the expression
		UNIVERSE,		// @ universe body (no ref)
		/* end of special bodies */
		ERR,			// . Error
		/* parenthesis */
		LEFT,			// (
		RIGHT,			// )
		/* prefix operators */
		//NOT,			// #	(complement operator)
		/* operators should be the last ones in the list */
		UNION,			// | or :
		PLUS,			// +
		MINUS,			// -	(also separator - terms of products)
	};

	enum NodeId : int {
		NodeId_Z  = 10,
		NodeId_X  = 11,
		NodeId_Y  = 12,
		NodeId_SZ = 13,
		NodeId_SX = 14,
		NodeId_SY = 15,
	};

protected:
	std::string	_name;	// name of body
	int	_id;		// index id in the bodies array
	Type	_type;		// FLUKA type of body
	dword	_color;		// body color
	int	_width;		// line width to use
	int	_generation;	// Generation number

	int	_nQ;		// Number of quads
	Quad	_Q[BODYQUADS];	// Surface quadratic equations

	//bool	_isexpanded;	// is gbody expanded
	//double _expand;	// scaling factor to expand body

	bool    _hasMatrix;	// if rotation matrix is set
	Matrix4	_matrix;	// rotation matrix
	Matrix4 _invMatrix;	// inverse rotation matrix

public:
	dword	show;		// display body (projection segments)
	Mesh*	_mesh;		// mesh of body

	// FIXME should be removed and placed to the Engine or the Cache
	//       GBody should not refer to higher level structures!
	// FIXME could be sorted list with pointer value for faster searching!!!
	Array<GZone*>	zones;	// zones that refer to this body

protected:
	Point	P, Po;		// position, outer position
	Vector	X, Y, Z;	// unit vectors along main axes
	double	xlen, ylen, zlen;// dimensions of X Y Z or radius

	// Save variables to be restored after a canceled operations (move/rotate/...)
	Type	stype;		// type
	Point	SP, SPo;	// positions
	Vector	SX, SY, SZ;	// vectors
	double	sxlen, sylen, szlen; // dimensions
	dword	sshow;		// show saved flag

	BBox::State _bboxState;	// status of bbox
	BBox	_bbox;		// cached information on bounding box
#ifdef USE_OBBOX
	BBox::State _obboxState;// status of obbox
	OBBox	_obbox[2];	// cached information on oriented bounding box
#endif

static	double	infinite;	// infinite dimension to use

public:
static	GNILBody	tnull;		// * empty/nil space
static	GUniverseBody	tuniverse;	// @ everything
static	GOPRBody	tleft;		// (
static	GOPRBody	tright;		// )
static	GOPRBody	tunion;		// |
static	GOPRBody	tplus;		// +
static	GOPRBody	tminus;		// -
//static	GOPRBody	tnot;	// # complement operator
static	const char*	_typeStr[];

public:
#if _DEBUG>0
static	int	verbose;	// enable verbose mode
static	bool	check;		// strict checking mode
#endif

public:
	GBody(const std::string& aname, const Type atype);
virtual	~GBody();

	// name
	void	name(const std::string& aname)	{ _name = aname; }
const	std::string&	name()		const	{ return _name; }
const	char*		cname()		const	{ return _name.c_str(); }

	// Id
	void	id(int i)		{ _id = i; }
	int	id()		const	{ return _id; }

	// Type
	Type	type()		const	{ return _type; }
const	char*	typeStr()	const	{ return _typeStr[(int)_type]; }

	// return name hash
	dword	hash()		const	{ return hash_djb2(cname()); }

	// normal body
	bool	isBody()	const	{ return _type <  Type::NIL; }
	bool	isAnyBody()	const	{ return _type <= Type::UNIVERSE; }
	bool	isRegion()	const	{ return _type == Type::REGION; }
	bool	isParenthesis()	const	{ return _type == Type::LEFT or _type == Type::RIGHT; }
	//bool	isAnyOperator()	const	{ return _type >= Type::NOT; }
	bool	isOperator()	const	{ return _type >= Type::UNION; }
	bool	isIntersection()const	{ return _type >= Type::PLUS;  }

	// FIXME maybe report also QUA with plane flag
	bool	isPlane()	const	{ return _type <= Type::P; }
	bool	isBox()		const	{ return In(_type, Type::RPP, Type::BOX); }
	bool	isSphere()	const	{ return InRange(Type::SPH, _type, Type::SZ); }
	bool	isInfCylinder() const	{ return InRange(Type::XCC, _type, Type::CZ); }
	bool	isQuad()	const	{ return In(_type, Type::QUA, Type::GQ, Type::SQ); }
	bool	isTorus()	const	{ return InRange(Type::TRX, _type, Type::TZ); }

	// Generation
	int	generation()	const	{ return _generation; }
	void	generation(int newGen) {
			_generation = newGen;
			resetBBox();
		}

	// get/set whas
virtual	void	set(int, double *) = 0;
virtual	int	get(double *)	const	{ return 0; }

	// Save/Restore position
virtual	void	save();
virtual	void	restore();
	Point	savedPosition()	const;

	// FIXME should be move to Engine or Cache
	// Zone reference
	void	addZone(GZone* zone)	{ zones.add(zone);   }
	void	delZone(GZone* zone)	{ zones.erase(zone); }
	bool	hasZone(GZone* zone)	{ return zones.find(zone)>=0; }

	// Position
	Point	position() const;
virtual	void	position(const Point& r);

	Vector	vectorX() const;
	Vector	vectorY() const;
	Vector	vectorZ() const;

	double	vectorXlen()	const	{ return xlen; }
	double	vectorYlen()	const	{ return ylen; }
	double	vectorZlen()	const	{ return zlen; }

	// Nodes
virtual	int	nodes()		const	{ return 1; }
virtual Point	node(const int)	const;

	// Mesh - create mesh if needed
	Mesh*	mesh() {
			if (_mesh==nullptr) {
				createMesh();
				if (_hasMatrix) _mesh->transform(_matrix);
			}
			return _mesh;
		}

	// Selection
virtual	int	closest(const Point& r, const double d, const Vector& w) const;

	// Matrix and transformations
	void	clearMatrix() {
				_hasMatrix = false;
				resetBBox();
			}
	bool	hasMatrix()	const	{ return _hasMatrix; }
	void	matrix(const Matrix4& M);
const	Matrix4& matrix()	const	{ return _matrix; }

	void	transform();
virtual void	move(int item, const Point& r, const Vector& w);
virtual	void	rotate(const double angle, const Vector& axis);

	// Quadrics
	int	nQ()		const	{ return _nQ; }
const	Quad&	Q(int n)	const	{ assert(n>=0 && n<_nQ); return _Q[n]; }

	// bounding box
	/** set user bbox */
	void	bbox(const BBox& bb);
	/** @return bounding box */
const	BBox&	bbox() const {
			assert(_bboxState != BBox::State::Invalid);
			return _bbox;
		}
#ifdef USE_OBBOX
const	OBBox&	obbox(bool in=false) const {
			assert(_obboxState!=BBox::State::Invalid);
			return _obbox[(int)in];
		}
#endif
	void	resetBBox() {
			_bboxState  = BBox::State::Invalid;
#ifdef USE_OBBOX
			_obboxState = BBox::State::Invalid;
#endif
		}

	/** update bounding boxes after any important change
	 * @param force	force recalculation of bounding box even if it is not invalid
	 */
	void	updateBBox(bool force=false) {
			if (force or _bboxState == BBox::State::Invalid) {
				_bbox.reset();
				_updateBBox();
				// FIXME it should be correct in the _updateBBox
				//if (_hasMatrix) _bbox.transform(_matrix);
				_bboxState = BBox::State::Valid;
			}
#ifdef USE_OBBOX
			if (force or _obboxState == BBox::State::Invalid) {
				_obbox[0].reset();
				_obbox[1].reset();
				_updateOBB(false);
				_updateOBB(true);
				//if (_hasMatrix) _obbox.transform(_matrix);	// FIXME
				_obboxState = BBox::State::Valid;
			}
#endif
		}

	// Location
virtual	Location locationWrt(const GBody* body) const;
	Location bbLocationWrt(const GBody* body) const;

	void	create(bool makeMesh=false) {
			createQuads();
			if (makeMesh) createMesh();
			if (hasMatrix()) transform();
		}

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
virtual	const char*	showX()	const { return "X"; }
virtual const char*	showY()	const { return "Y"; }
virtual const char*	showZ()	const { return "Z"; }

	Vector	normal(const double x, const double y, const double z) const;
	Vector	normal(const Point& p) const	{ return normal(p.x, p.y, p.z); }

	bool	inside(	const double x, const double y, const double z,
			const double u, const double v, const double w,
			const double* acc=NULL,
#ifdef EXPERIMENTAL
			bool* withinAcc=NULL,
#endif
			const int ignore1=-1, const int ignore2=-1) const;

	Side	side(const double x, const double y, const double z,
			const double dsurf=SMALL) const;
	bool	inside( const double x, const double y, const double z,
			const Quad* ignore_a, const Quad* ignore_b, const Quad* ignore_c) const;
	bool	outside(const double x, const double y, const double z,
			const Quad* ignore_a, const Quad* ignore_b, const Quad* ignore_c) const;

	/** return nearest (safe) distance in any direction to exit (or enter) body
	 * @param x,y,z	starting position
	 * @return safe distance
	 */
	// FIXME populate on body types, right now is using the generic one
	// which is not optimum
virtual	double	distance(const double x, const double y, const double z) const;
	double	distance(const Point& p) const	{ return distance(p.x, p.y, p.z); }

	/** return distance of intersection (tmin) of a ray starting from (x,y,z) with
	 * direction (u,v,w) from in the range (tmin..tmax] distance.
	 * @param x,y,z	starting position
	 * @param u,v,w	direction of ray/particle
	 * @param tmin	I/O starting distance, and on exit contains the returned distance
	 * @param tmax	maximum distance to check
	 *
	 * @warning not cached
	 */
/*virtual*/	bool	distance(const double x, const double y, const double z,
			         const double u, const double v, const double w,
			         double* tmin, double* tmax) const;
			 //double& t[]) const;	// for multiple solutions of a generalized body

	bool	distance(const Point& p, const Vector& v, double* tmin, double* tmax) const
			{ return distance(p.x,p.y,p.z, v.x,v.y,v.z, tmin, tmax); }

	/** return distance of intersection (tmin) of a ray starting from (x,y,z) with
	 * direction (u,v,w) from in the range (tmin..tmax] distance.
	 * @param x,y,z	starting position
	 * @param u,v,w	direction of ray/particle
	 * @param tmin	I/O starting distance, and on exit contains the returned distance
	 * @param tmax	maximum distance to check
	 * @param cache	body caching array
	 *
	 * @return true  if body is convex  with solution  [tmin..tmax]
	 *         false if body is concave with solutions (-inf..tmin] | [tmax..+inf)
	 *
	 * @warning body caching has to be initialized (resized) to the max number of bodies
	 *          and the cache.inc() has to be called when ever the position (x,y,z)
	 *          or the direction (u,v,w) has changed.
	 */
	bool	distance(const double x, const double y, const double z,
			 const double u, const double v, const double w,
			 double* tmin,   double* tmax,
			 Cache<CBody3D>& cache) const {
				assert(!cache.empty());
				CBody3D& cbody = cache[id()];
				if (cache.isOk(cbody)) {
					*tmin = cbody.tmin;
					*tmax = cbody.tmax;
#ifdef STRONGDEBUG
					cbody._assert(x,y,z, u,v,w);
#endif
					return cbody.tinv;
				}
				cbody.tinv = distance(x, y, z, u, v, w, tmin, tmax);
				cbody.tmin = *tmin;
				cbody.tmax = *tmax;
				cache.mark(cbody);
#ifdef STRONGDEBUG
				cbody.store(x,y,z, u,v,w);
#endif
				return cbody.tinv;
			}

	bool	distance(const Point& p, const Vector& v, double* tmin, double* tmax,
			 Cache<CBody3D>& cache) const {
				return distance(p.x,p.y,p.z, v.x,v.y,v.z, tmin, tmax, cache);
			}

	/**
	 * Find if a point (x,y,z)+t*(u,v,w) is inside the body or not, using the
	 * information from the distance
	 * Returns: true  inside
	 *          false outside
	 */
	//bool	inside( const double x, const double y, const double z,
	//		const double u, const double v, const double w,
	//		const double t) const {
	//			double tmin, tmax;
	//			bool tinv = distance(x,y,z, u,v,w, &tmin,&tmax);
	//			if (tinv)
	//				return (bool)(t<tmin or tmax<=t);
	//			else
	//				return (bool)(tmin<=t and t<tmax);
	//		}
	//bool	inside(const Vector& p, const Vector &d, const double t) const
	//		{ return inside(p.x,p.y,p.z, d.x,d.y,d.z, t); }

	/** check is the point defined as (x+u*t,y+v*t,z+w*t) towards (u,v,w) at distance t
	 * is inside or outside of current body
	 * If on boundary the direction is considered outside if the dot product
	 * of the normal and direction is positive
	 *
	 * @param x,y,z	starting position of ray/particle
	 * @param u,v,w	direction of ray/particle
	 * @param t	distance along (u,v,w) direction
	 *
	 * @return true if inside, false otherwise
	 */
	bool	inside( const double x, const double y, const double z,
			const double u, const double v, const double w,
			const double t,
			Cache<CBody3D>& cache) const {
				assert(!cache.empty());
				CBody3D& cbody = cache[id()];
				if (not cache.isOk(cbody)) {
					cbody.tinv = distance(x,y,z, u,v,w, &cbody.tmin,&cbody.tmax);
					cache.mark(cbody);
#ifdef STRONGDEBUG
					cbody.store(x,y,z, u,v,w);
				} else {
					cbody._assert(x,y,z, u,v,w);
#endif
				}
				if (cbody.tinv)
					return (bool)(t<cbody.tmin or cbody.tmax<=t);
				else
					return (bool)(cbody.tmin<=t and t<cbody.tmax);
			}
	bool	inside( const Point& p, const Vector& v, const double t,
			Cache<CBody3D>& cache) const
			{ return inside(p.x,p.y,p.z, v.x,v.y,v.z, t, cache); }

	// Drawing options
	// Color
	void	color(dword c)		{ _color = c & FLAG_COLORMASK; }
	dword	color() const		{ return _color; }

	// line width
	void	lineWidth(int w)	{ _width = w; }
	int	lineWidth() const	{ return _width; }

	size_t	memory() const;

protected:
	void	checkOrthogonal() const;

virtual	void	createQuads()		{ _nQ = 0; }
virtual	void	createMesh();

virtual	void	_updateBBox() = 0;
#ifdef USE_OBBOX
virtual	void	_updateOBB(bool in=false) = 0;
#endif

virtual	Location _locationWrt(const GBody* body) const;

	// add a new quad
	void	addQuad(const Quad& q);
	void	addQuad(double Cxx, double Cyy, double Czz,
			double Cxy, double Cxz, double Cyz,
			double Cx,  double Cy,  double Cz,
			double C);
	void	addQuad(double Cx,  double Cy,  double Cz, double C);

	void	findXYZ();

	void	roundFloat(double& f) const	{ if (Eq0(f,SMALL3)) f = 0.0; }
	void	roundVector(Vector& v) const {
			if (Eq0(v.x,SMALL3)) v.x = 0.0;
			if (Eq0(v.y,SMALL3)) v.y = 0.0;
			if (Eq0(v.z,SMALL3)) v.z = 0.0;
		}

	void	round0(double* what, int n, const double eps) const;
	//void	setEdge(const int i, const int a, const int b) { edge(i).set(&vertex(a), &vertex(b)); }

public:
static	GBody*	newBody(const std::string& name, const std::string&  type);
static	GBody*	newBody(const std::string& name, const Type type);
static	int	compare(GBody* const& a, GBody* const& b)
			{ return strcmp(a->cname(),b->cname()); }
static	const char* typeStr(Type t)		{ return _typeStr[(int)t]; }
}; // GBody

// ================================= GNILBody ==================================
class GNILBody : public GBody {
public:
	GNILBody() : GBody("*", Type::NIL) {}

	virtual	void	set(int, double *)	{};
protected:
	// bounding box
	virtual	void	_updateBBox()		{ _bbox.null(); };
#ifdef USE_OBBOX
	virtual	void	_updateOBB(bool)	{ throw 0; };
#endif
	virtual	Location _locationWrt(const GBody *) const { return Location::OUTSIDE; };
}; // GNILBody

// =============================== GUniverseBody ===============================
class GUniverseBody : public GBody {
public:
	GUniverseBody() : GBody("@", Type::UNIVERSE) {}

	virtual	void	set(int, double *)	{};
protected:
	// bounding box
	virtual	void	_updateBBox()		{ _bbox.infinite(); };
#ifdef USE_OBBOX
	virtual	void	_updateOBB(bool)	{ throw 0; };
#endif
	virtual	Location _locationWrt(const GBody *) const { return Location::BinA; };
}; // GUniverseBody

// ================================= GOPRBody ==================================
/** Operator body null, +, -, |, (, ) */
class GOPRBody : public GBody {
public:
	GOPRBody(const std::string& aname, const Type atype)
		: GBody(aname, atype) {}

	virtual	void	set(int, double *)	{};
	virtual	void	position(const Point&)	{};
	virtual	int	nodes()		const	{ throw 0; }
	virtual void	move(int, const Point&, const Vector&) {};
	virtual	void	rotate(const double, const Vector&) {};

protected:
	// bounding box
	virtual	void	_updateBBox()		{ throw 0; };
#ifdef USE_OBBOX
	virtual	void	_updateOBB(bool)	{ throw 0; };
#endif
	virtual	Location _locationWrt(const GBody *) const { throw 0; };
}; // GOPRBody

// ================================= GERRBody ==================================
/* Error body */
class GERRBody : public GBody {
public:
	GERRBody(const std::string& aname, const Type atype)
		: GBody(aname, atype) {}

	virtual	void	set(int, double *)	{};
	virtual	void	position(const Point&)	{};
	virtual	int	nodes()		const	{ throw 0; }
	virtual void	move(int, const Point&, const Vector&) {};
	virtual	void	rotate(const double, const Vector&) {};

protected:
	// bounding box
	virtual	void	_updateBBox()		{ throw 0; };
#ifdef USE_OBBOX
	virtual	void	_updateOBB(bool)	{ throw 0; };
#endif
	virtual	Location _locationWrt(const GBody *) const { throw 0; };
}; // GERRBody

// ============================== GBodyOrderAccel ==============================
/** GBodyOrderAccel
 * Used by intersectRayUndefinedRegion to maintain a list of bodies ordered by its
 * tmin and tmax distance to the ray origin
 */
class GBodyOrderAccel {
public:
	GBody* body;
	double t;
public:
	GBodyOrderAccel() : body(NULL), t(-INFINITE) {}
	GBodyOrderAccel(GBody* ab, double at) : body(ab), t(at) {}
}; // GBodyOrderAccel

std::ostream& operator << (std::ostream&, const GBody&);
#endif
