/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#ifndef __GEOMETRY_LIGHT_H
#define __GEOMETRY_LIGHT_H

#include "math/eps.h"
#include "math/point.h"

// =================================== Light ===================================
class Light {
public:
	enum class Type {
		Sun,		// Sun light, uniform directional light
		Omni,		// Omni light up to a certain radius
		Spot		// Spot light
	};

	enum class FallOff : int {
		Constant  = 0,	// Constant power over distance
		Inverse   = 1,	// power ~= 1/distance
		Inverse2  = 2	// power ~= 1/distance**2
	};

	Type	type;		// light type
	bool	relative;	// relative or absolute position
	Point	pos;		// absolute position
	Vector	dir;		// absolute direction
	FallOff	falloff;	// light fall off type
				//	 0 = constant
				//	 1 = inverse linear
				//	 2 = inverse square
	double	distance;	// maximum distance = |dir|
	double	power;		// light power
	double  radius;		// volumetric size (default = 0.0 for point lights)
	int	nsamples;	// nr. of light samples used for volumetric lights
	bool	spec;		// make specular shine
	bool	shadow;		// cast shadows

public:
	Light() : type(Type::Sun),
		  relative(false),
		  pos(0,0,0),
		  dir(0,0,1),
		  falloff(FallOff::Constant),
		  distance(INFINITE),
		  power(1.0),
		  radius(0),
		  nsamples(1),
		  spec(false),
		  shadow(false)
		{}
	Light(const Light& p) = default;
	Light& operator = (const Light& p) = default;
}; // class Light

// ================================== VLight ===================================
// Viewer specific light, to update the position, direction in case of relative
// light
struct VLight {
	Point	 pos;		// relative position in viewport
	Vector	 dir;		// relative direction in viewport
}; // struct VLight

#endif
