/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "planebody.h"

#include "mesh.h"
#include "cylbody.h"
#include "spherebody.h"
#include "math/point.h"
#include "math/vector.h"

// ================================= GPLABody ==================================
/** set */
void GPLABody::set( const double nx, const double ny, const double nz,
		    const double  x, const double  y, const double  z)
{
	assert(type()==Type::PLA);
	Z.set(nx, ny, nz);
	P.set(x, y, z);
	zlen = Z.normalize();
	if (Eq0(zlen,SMALL)) {
		Z = Vector::Zo;
		zlen = 1.0;
		throw std::string("PLA normal is null");
	}
	findXYZ();
} // set

/** set */
void GPLABody::set(const double pos)
{
	switch (type()) {
		case Type::YZP:
		case Type::PX:
			Z = Vector::Xo;
			X = Vector::Yo;
			Y = Vector::Zo;
			P.set(pos, 0.0, 0.0);
			break;

		case Type::XZP:
		case Type::PY:
			Z = Vector::Yo;
			X = Vector::Zo;
			Y = Vector::Xo;
			P.set(0.0, pos, 0.0);
			break;

		case Type::XYP:
		case Type::PZ:
			Z = Vector::Zo;
			X = Vector::Xo;
			Y = Vector::Yo;
			P.set(0.0, 0.0, pos);
			break;

		default:
			assert(false);
	}
} // set

/** set */
void GPLABody::set(int /*n*/, double* what)
{
	if (type()==Type::PLA) {
		round0(what, 6, SMALL);
		set(what[0], what[1], what[2], what[3], what[4], what[5]);
	} else
	if (type() == Type::P) {
		round0(what, 4, SMALL);
		Z.set(what[0], what[1], what[2]);
		double z2 = Z.length2();
		if (z2>SMALL) {
			double t = what[3] / z2;
			P.set(Z.x*t, Z.y*t, Z.z*t);
			Z /= sqrt(z2);
		} else {
			Z = Vector::Zo;
			zlen = 1.0;
			throw std::string("P normal is null");
		}
		findXYZ();
	} else {
		round0(what, 1, SMALL);
		set(what[0]);
	}
} // set

/** get */
int GPLABody::get(double* what) const
{
	switch (type()) {
		case Type::PLA:
			what[0] = zlen*Z.x;
			what[1] = zlen*Z.y;
			what[2] = zlen*Z.z;
			what[3] = P.x;
			what[4] = P.y;
			what[5] = P.z;
			return 6;

		case Type::P:
			what[0] = zlen*Z.x;
			what[1] = zlen*Z.y;
			what[2] = zlen*Z.z;
			what[3] = Z*P;
			return 4;

		case Type::YZP:
		case Type::PX:
			what[0] = P.x;
			return 1;

		case Type::XZP:
		case Type::PY:
			what[0] = P.y;
			return 1;

		case Type::XYP:
		case Type::PZ:
			what[0] = P.z;
			return 1;

		default:
			return -1;
	}
} // get

/** position
 * @param x,y,z	new position to set
 */
void GPLABody::position(const Point& r)
{
	GBody::position(r);

	switch (type()) {
		case Type::YZP:
		case Type::PX:
			P.y = P.z = 0.0;
			break;

		case Type::XZP:
		case Type::PY:
			P.x = P.z = 0.0;
			break;

		case Type::XYP:
		case Type::PZ:
			P.x = P.y = 0.0;
			break;

		default: NOP;
	}
	Po = P + zlen*Z;
} // position

/** return sign of the w axis that has to be used for move=20 type
 * see below on move
 */
void GPLABody::signMove20(const Point& r, const Vector& w)
{
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= P;
	Vector dw = T ^ w;
	_sign20 = Sign(dw*Z);
} // signMove20

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 * WARNING: the w is used for the rotation (20)
 *          and should be either viewport w or -w such as
 *                    ((r-p)^w)*Z > 0
 *          The reason is that the system below gives two solutions
 *          depending on the side that the user start to drag the extremity
 *          and the program has to remember the side
 */
void GPLABody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= SP;
	if (item==NodeId_axisR) {
		// Rotate plane using r as a Second node lying on the plane
		// Calculate new normal
		// Vector w = SZ ^ T;
		// Z = T ^ w;

		// We have to rotate the plane around w of the viewport
		// keeping the dot(W,N')=const
		// Equations to fulfill
		// 1. N'.r - N'.p = 0 <=> N'*T = 0
		// 2. W.N' = W.N
		// 3. |N'| = |N|

		Point&  d = T;
		Vector dw = d ^ w;
		double  c = SZ * w;

		double E = d.x*dw.z;
		if (!Eq0(E,SMALL)) {
			double A = c*d.x*d.y;
			double B = d.y*dw.y + d.z*dw.z;
			double C = c*d.x*d.x;
			double D = d.x*dw.y;

			double AA = B*B + D*D + E*E;
			double BB = 2.0 * (A*B + C*D);
			double CC = A*A + C*C - E*E;

			if (Eq0(AA,SMALL)) return;
			double x1, x2;
			int n = quadratic(BB/AA, CC/AA, &x1, &x2, SMALL);
			if (n==0) return;
			Z.z = x1;
			Z.y =  (C + D*Z.z) / E;
			Z.x = -(A + B*Z.z) / E;

			if (Sign(dw * Z)*_sign20 < 0) {
				Z.z = x2;
				Z.y =  (C + D*Z.z) / E;
				Z.x = -(A + B*Z.z) / E;
			}
		} else {
			// rotate x->y->z->x
			E = d.y*dw.x;
			if (!Eq0(E,SMALL)) {
				double A = c*d.y*d.z;
				double B = d.z*dw.z + d.x*dw.x;
				double C = c*d.y*d.y;
				double D = d.y*dw.z;

				double AA = B*B + D*D + E*E;
				double BB = 2.0 * (A*B + C*D);
				double CC = A*A + C*C - E*E;

				if (Eq0(AA,SMALL)) return;
				double x1, x2;
				int n = quadratic(BB/AA, CC/AA, &x1, &x2, SMALL);
				if (n==0) return;
				Z.x = x1;
				Z.z =  (C + D*Z.x) / E;
				Z.y = -(A + B*Z.x) / E;

				if (Sign(dw * Z)*_sign20 < 0) {
					Z.x = x2;
					Z.z =  (C + D*Z.x) / E;
					Z.y = -(A + B*Z.x) / E;
				}
			} else {
				// rotate x->z, y->x, z->y
				E = d.z*dw.y;
				if (Eq0(E,SMALL)) return;
				double A = c*d.z*d.x;
				double B = d.x*dw.x + d.y*dw.y;
				double C = c*d.z*d.z;
				double D = d.z*dw.x;

				double AA = B*B + D*D + E*E;
				double BB = 2.0 * (A*B + C*D);
				double CC = A*A + C*C - E*E;

				if (Eq0(AA,SMALL)) return;
				double x1, x2;
				int n = quadratic(BB/AA, CC/AA, &x1, &x2, SMALL);
				if (n==0) return;
				Z.y = x1;
				Z.x =  (C + D*Z.y) / E;
				Z.z = -(A + B*Z.y) / E;

				if (Sign(dw * Z)*_sign20 < 0) {
					Z.y = x2;
					Z.x =  (C + D*Z.y) / E;
					Z.z = -(A + B*Z.y) / E;
				}
			}
		}
		assert(Eq(Z*w,c,SMALL));

		Z.normalize();
		try {
			findXYZ();
		} catch (std::string& /*e*/) {
			/* do nothing) */
		}
	} else
	if (item==NodeId_Pos)
		P = SP + (T*SZ)*SZ;
	else
		GBody::move(item, r, w);
	checkType();
} // move

/** rotate body around an axis */
void GPLABody::rotate(const double angle, const Vector& axis)
{
	GBody::rotate(angle,axis);
	checkType();
} // rotate

/** check plane type */
void GPLABody::checkType()
{
	// Check type
	switch (Z.direction(SMALL)) {
		case Vector::Direction::X:
		case Vector::Direction::Xn:
			if (type()<=Type::PLA)		// FLUKA
				_type = Type::YZP;
			else
				_type = Type::PX;
			break;
		case Vector::Direction::Y:
		case Vector::Direction::Yn:
			if (type()<=Type::PLA)		// FLUKA
				_type = Type::XZP;
			else
				_type = Type::PY;
			break;
		case Vector::Direction::Z:
		case Vector::Direction::Zn:
			if (type()<=Type::PLA)		// FLUKA
				_type = Type::XYP;
			else
				_type = Type::PZ;
			break;
		default:
			if (type()<=Type::PLA)		// FLUKA
				_type = Type::PLA;
			else
				_type = Type::P;
	}
} // checkType

/** createQuads */
void GPLABody::createQuads()
{
	GBody::createQuads();
	//      Cx     Cy     Cz      C
	addQuad(Z.x, Z.y, Z.z, -Z*P);
} // createQuads

/* createMesh */
void GPLABody::createMesh()
{
	GBody::createMesh();
	if (_mesh->isEmpty())
		_mesh->createPlane(P, X, xlen, Y, ylen, Z, zlen);
	else
		_mesh->updatePlane(P, X, xlen, Y, ylen, Z, zlen);
} // createMesh

/** @return bounding box of body */
void GPLABody::_updateBBox()
{
	_bbox.infinite();

	Point p = GBody::position();

	switch (vectorZ().direction(SMALL)) {
		case Vector::Direction::Xn:
			_bbox.lowx(p.x);
			break;

		case Vector::Direction::X:
			_bbox.highx(p.x);
			break;

		case Vector::Direction::Yn:
			_bbox.lowy(p.y);
			break;

		case Vector::Direction::Y:
			_bbox.highy(p.y);
			break;

		case Vector::Direction::Zn:
			_bbox.lowz(p.z);
			break;

		case Vector::Direction::Z:
			_bbox.highz(p.z);
			break;

		default:
			/*do nothing - infinite*/;
	}
} // _updateBBox

#ifdef USE_OBBOX
/** @return oriented bounding box of body */
void GPLABody::_updateOBB(bool in)
{
	OBBox& bb = _obbox[(int)in];
	bb.infinite();

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();
	bb.highPt.z = 0.0;
} // _updateOBB
#endif

/**
 * @param body	other body to check location against
 * @return location of this body with respect to another body
 *
 * FIXME TODO add bounding box limits!!!!
 */
GBody::Location GPLABody::_locationWrt(const GBody* body) const
{
	if (body->isPlane()) {
		// Check if normals are co-linear
		Vector thisN = vectorZ();
		Vector bodyN = body->vectorZ();

		double ptot2 = thisN.length2()*bodyN.length2();
		if (ptot2<=0.0) return Location::OVERLAP;	// error
		double dot = (thisN*bodyN)/sqrt(ptot2);

		// if |dot|!=1 they will overlap somewhere
		if (Eq(dot,  1.0, SMALL)) {
			// Normals are parallel
			// Two situations AinB or BinA
			// Check if this is inside body or vice versa
			double pos = Q(0)(body->position());
			if (pos < 0.0)
				return Location::BinA;
			else
				return Location::AinB;
		} else
		if (Eq(dot, -1.0, SMALL)) {
			// Normals are anti-parallel
			// Two situations Overlap or Outside
			double pos = Q(0)(body->position());
			if (pos < 0.0)
				// FIXME could be optimized in the level of Zone::optimize
				// but here we can only return an OVERLAP
				return Location::OVERLAP;
			else
				return Location::OUTSIDE;
		} else
			return Location::OVERLAP;

	} else
	if (body->isSphere()) {
		// Check distance of SPH center to plane vs radius
		const Quad& q = Q(0);
		double d = q(body->position()) / sqrt(Sq(q.Cx) + Sq(q.Cy) + Sq(q.Cz));
		double r = ((const GSPHBody*)body)->R();
		if (d > r)
			return Location::OUTSIDE;
		else
		if (d < -r)
			return Location::BinA;
		else
			return Location::OVERLAP;
	} else
	if (body->isBox() or body->type()==Type::WED) {
		double n = sqrt(Sq(Q(0).Cx) + Sq(Q(0).Cy) + Sq(Q(0).Cz));
		Point  bP = body->position();
		Vector bX = body->vectorX() * body->vectorXlen();
		Vector bY = body->vectorY() * body->vectorYlen();
		Vector bZ = body->vectorZ() * body->vectorZlen();
		Point V[8] = {  bP,       bP+bX,    bP+bY,
				bP+bZ,    bP+bX+bZ, bP+bY+bZ,
				bP+bX+bY, bP+bX+bY+bZ };
		size_t nV = body->type()==Type::WED? 6 : 8;
		double d = Q(0)(V[0]) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		bool ins = d < 0.0;
		for (size_t i=1; i<nV; i++) {
			d = Q(0)(V[i]) / n;
			if (Eq0(d, SMALL)) return Location::OVERLAP;
			if (ins) {
				if (d > 0.0) return Location::OVERLAP;
			} else
				if (d < 0.0) return Location::OVERLAP;
		}
		if (ins)
			return Location::BinA;
		else
			return Location::OUTSIDE;
	} else
	if (body->type()==Type::ARB) {
		// FIXME normally this check could be done for every object
		// assuming a safe sagitta distance
		assert(_mesh!=nullptr);
		double n = sqrt(Sq(Q(0).Cx) + Sq(Q(0).Cy) + Sq(Q(0).Cz));
		double d = Q(0)(body->_mesh->vertex(0)) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		bool ins = d < 0.0;
		for (size_t i=1; i<body->_mesh->nvertices(); i++) {
			d = Q(0)(body->_mesh->vertex(i)) / n;
			if (Eq0(d, SMALL)) return Location::OVERLAP;
			if (ins) {
				if (d > 0.0) return Location::OVERLAP;
			} else
				if (d < 0.0) return Location::OVERLAP;
		}
		if (ins)
			return Location::BinA;
		else
			return Location::OUTSIDE;
	} else
	if (body->isIntersection()) {
		// Verify if axis of cylinder is perpendicular to plane
		Vector thisN = vectorZ();
		Vector bodyZ = body->vectorZ();
		if (Eq0(thisN * bodyZ, SMALL)) {
			// Check distance of cylinder axis to plane
			const Quad& q = Q(0);
			double d = q(body->position()) / sqrt(Sq(q.Cx) + Sq(q.Cy) + Sq(q.Cz));
			double r = ((const GInfEllCylBody*)body)->R();
			if (d > r)
				return Location::OUTSIDE;
			else
			if (d < -r)
				return Location::BinA;
			else
				return Location::OVERLAP;
		} else
			return Location::OVERLAP;
	//} else
	//if (body->type() == Type::QUA) {
	//	Plane: ax + by + cz + d = 0
	//	Substituting z = - (ax+by+d)/c  (if c!=0) to QUA
	//	we get a Conic equation on QUA(x,y)
	//	Check if Conic is degenerate or not!
	} else {
		BBox bb = body->bbox();
		double n = sqrt(Sq(Q(0).Cx) + Sq(Q(0).Cy) + Sq(Q(0).Cz));

		double d = Q(0)(bb.low()) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		bool ins = d < 0.0;

		Point p(bb.low().x, bb.low().y, bb.high().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.low().x, bb.high().y, bb.high().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.low().x, bb.high().y, bb.low().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.high().x, bb.low().y, bb.low().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.high().x, bb.low().y, bb.high().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.high().x, bb.high().y, bb.high().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.high().x, bb.high().y, bb.low().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		if (ins)
			return Location::BinA;
		else
			return Location::OUTSIDE;
	}
	return GBody::_locationWrt(body);
} // _locationWrt

// ================================== GP3Body ==================================
/** set */
void GP3Body::set(const Point& p1, const Point& p2, const Point& p3)
{
	P  = p1;
	PB = p2 - P;
	PC = p3 - P;
	update();
} // set

/** update plane parameters from the 3 points */
void GP3Body::update()
{
	Z = PB ^ PC;
	zlen = Z.normalize();
	double d = Z * P;
	// From MCNP4B manual 3-18			// Ax+By+Cz+D=0
	if (Eq0(d,1e-10)) {				// D=0 passes from the origin
		if (Eq0(Z[2],1e-10) and Eq0(Z[1],1e-10)) // B=C=D=0  (Ax=0)
			Z = Vector::Xo;			// Z*(inf,0,0) is positive => Z.x=+1
		else
		if (Eq0(Z[2],1e-10))			// C=D=0  (Ax+By=0)
			Z.y = Abs(Z.y);			// Z*(0,inf,0) is positive => Z.y=|Z.y|
		else					// D=0
			Z.z = Abs(Z.z);			// Z*(0,0,inf) is positive => Z.z=|Z.z|
	} else
	if (d < 0.0)
		Z = -Z;
	findXYZ();
} // update

/** set */
void GP3Body::set(int /*n*/, double* what)
{
	round0(what, 9, SMALL);
	set(	Point(what[0], what[1], what[2]),
		Point(what[3], what[4], what[5]),
		Point(what[6], what[7], what[8]));
} // set

/** get */
int GP3Body::get(double* what) const
{
	what[0] = P.x;
	what[1] = P.y;
	what[2] = P.z;
	what[3] = PB.x + P.x;
	what[4] = PB.y + P.y;
	what[5] = PB.z + P.z;
	what[6] = PC.x + P.x;
	what[7] = PC.y + P.y;
	what[8] = PC.z + P.z;
	return 9;
} // get

/** save */
void GP3Body::save()
{
	SPB = PB;
	SPC = PC;
	GBody::save();
} // save

/** restore */
void GP3Body::restore()
{
	PB = SPB;
	PC = SPC;
	GBody::restore();
} // restore

/** @return transformed position */
Point GP3Body::node(const int n) const
{
	Point r;
	switch (n) {
		case 0: r = P;
			break;
		case 1: r = P+PB;
			break;
		case 2: r = P+PC;
			break;
	}
	return _hasMatrix? _matrix*r : r;
} // node

/**  @return closest point, edge, face */
int GP3Body::closest(const Point& r, const double d, const Vector& w) const
{
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= P;
	if (T.isClose(PB, d)) return NodeId_Y;
	if (T.isClose(PC, d)) return NodeId_Z;
	return GBody::closest(r,d,w);
} // closest

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GP3Body::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= SP;
	switch (item) {
		case 1:
			P = SP + (T*Z)*SZ;
			roundVector(P);
			break;
		case NodeId_Y:
			PB = T;
			roundVector(PB);
			update();
			break;
		case NodeId_Z:
			PC = T;
			roundVector(PC);
			update();
			break;
		default:
			GBody::move(item, r, w);
	}
} // move
