/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	11-Oct-2016
 */

#include "tetrameshio.h"

#include <assert.h>

#include "vertex.h"
#include "tetramesh.h"
#include "tool/token.h"

// ================================ TetraMeshIO ================================
using std::cout;
using std::endl;

/** read a mesh from an MESH file, ASCII or binary
 * @param mesh to populate
 * @return true if everything ok
 */
bool TetraMeshIO::read(TetraMesh& mesh)
{
	char dummy[12];
	if (!open(Stream::IO::Read, false)) return false;
	stream.read(dummy, 11); dummy[11] = 0;
	if (!strcmp(dummy, "$MeshFormat"))
		return _readAscii(mesh);
	else {
		close();
		open(Stream::IO::Read, true);
		return _readBinary(mesh);
	}
} // read

/** read mesh from acsii .msh file
 * @return true on success, false otherwise
 */
bool TetraMeshIO::_readAscii(TetraMesh& mesh)
{
//	char buf[256];

	stream.seekg(0, std::ifstream::beg);		 // go to start

	Token token(stream);
	token.ignore_newline = true;
	token.setAlpha('$');
	token.next();

	// Could be that we have many solids.... (not implemented at the moment)
	int datasize, filetype;
	int n;
	double version;

	if (!token.mustbe("$MeshFormat","$MeshFormat expected")) goto _ERROR;

	version = token.getNumber();
	filetype = token.getInteger();
	datasize = token.getInteger();

	cout << "Version=" << version << " Type=" << filetype << " size=" << datasize << endl;
	if (!token.mustbe("$EndMeshFormat","$EndMeshFormat expected")) goto _ERROR;

	if (!token.mustbe("$Nodes","$Nodes expected")) goto _ERROR;
	n = token.getInteger();
	for (int i=0; i<n; i++) {
		int j = token.getInteger();
		assert(i+1 == j);
		if (i+1 != j) goto _ERROR;
		Vertex v;
		v.x = token.getNumber();
		v.y = token.getNumber();
		v.z = token.getNumber();
		mesh.add(v);
		// FIXME how to ensure that we don't have duplicates!
	}
	if (!token.mustbe("$EndNodes","$EndNodes expected")) goto _ERROR;
	if (!token.mustbe("$Elements","$Elements expected")) goto _ERROR;
	n = token.getInteger();
	for (int i=0; i<n; i++) {
		int j = token.getInteger();
		assert(i+1 == j);
		if (i+1 != j) goto _ERROR;
		int etype  = token.getInteger();
		int ntags  = token.getInteger();
		/*int entity =*/ token.getInteger();
		// skip remaining tags
		for (int k=1; k<ntags; k++) token.getInteger();

		if (etype==4) {
			// tetrahedron
			int a = token.getInteger()-1;
			int b = token.getInteger()-1;
			int c = token.getInteger()-1;
			int d = token.getInteger()-1;
			mesh.add(a,b,c,d);
		} else {
			token.skipLine();
		}
	}
	if (!token.mustbe("$EndElements","$EndElements expected")) goto _ERROR;

#if 0
//	mesh.process();
#endif

	close();
	return true;

_ERROR:
	close();
	return false;
} // _readAscii

/** Read mesh form binary gmsh file */
/** read mesh from binary stl.file
 * @return true on success, false otherwise
 */
bool TetraMeshIO::_readBinary(TetraMesh&)
{
	close();
	return false;
} // _readBinary
