/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	05-Apr-2011
 *
 * TODO
 * GObject has to be split in two GObject and VObject
 * otherwise we will have problems when parallel drawings are done
 */

#ifndef __GOBJECT_H
#define __GOBJECT_H

#include <Python.h>

#include <X11/X.h>
#include <X11/Xutil.h>

#include "os.h"
#include "math/bbox.h"
#include "math/bmath.h"
#include "math/spline.h"
#include "geometry/geo.h"
#include "geometry/geometry.h"
#include "viewer/viewer.h"

struct ViewerObject;

enum ClassType {
	GObjectClass  = 0,   // option offset also
	GPointClass   = 10,
	GArrowClass   = 20,
	GRulerClass   = 30,
	GRotdefiClass = 40,
	GLightClass   = 50,
	GBeamClass    = 60,
	GCameraClass  = 70,
	GSplineClass  = 100, // allow for up to 10 points
	GMeshClass    = 200
};

// Text location
enum AnchorType {
	Anchor_none,	// Do not display text
	Anchor_C,	// Center
	Anchor_N,	// North
	Anchor_NE,	// North-East
	Anchor_E,	// East
	Anchor_SE,	// South-East
	Anchor_S,	// South
	Anchor_SW,	// South-West
	Anchor_W,	// West
	Anchor_NW	// North-West
};

// Object type
enum ObjectType {
	Object_Default     = 0,

	Point_Cross        = GPointClass,
	Point_Dot          = GPointClass+1,
	Point_Square       = GPointClass+2,
	Point_X            = GPointClass+3,
	Point_Diamond      = GPointClass+4,
	Point_Circle       = GPointClass+5,

	Arrow_Head         = GArrowClass,
	Arrow_Tail         = GArrowClass+1,
	Arrow_HeadTail     = GArrowClass+2,
	Arrow_Line         = GArrowClass+3,

	Ruler_Simple       = GRulerClass,
	Ruler_Angle        = GRulerClass+1,

	Rotdefi_XYZ        = GRotdefiClass,

	Light_Sun          = GLightClass,
	Light_Omni         = GLightClass+1,
	Light_Spot         = GLightClass+2,

	Beam_Arrow         = GBeamClass,
	Beam_Divergence    = GBeamClass+1,

	Camera_Persp       = GCameraClass,
	Camera_Ortho       = GCameraClass+1,

	Spline_Open        = GSplineClass,
	Spline_Closed      = GSplineClass|1,
	Spline_Lines       = GSplineClass+2,
	Spline_LinesClosed = (GSplineClass+2)|1,

	Mesh_Default       = GMeshClass,
};

// ================================== GObject ==================================
/** GObject graphics object class */
class GObject {
public:
	enum NodeId : int {
		NodeId_1   = 1,
		NodeId_2   = 2,
	};
public:
	ObjectType	type;		// object type to draw
	AnchorType	anchor;		// text location
	dword		color;		// drawing color
	bool		select;		// selected
	int		selectNode;	// selected node
					// 0=object is selected, 1=first node, 2=second.
	bool		show;		// show object
	double		drawDw;		// visibility limit (furstrum in w-axis)
	int		size;		// object dimension
	int		lineWidth;	// line width to use

protected:
	std::string	_name;		// name of object
	int		_id;		// id of object
	bool		_visible;	// if object is visible in viewport
	Point		P;		// absolute position

	Point		VP, VPc;	// viewer position, and clipped
	int		xpos, ypos;	// object position

	Point		SP;		// save position

	XGCValues	gcValues;	// context values
//	Font		font;

public:
static	dword		nodeColor;	// selection color for nodes
static	dword		selectColor;	// if !=0 then is selected and contains the selection color

public:
	GObject(const std::string& aname="", ObjectType o=Object_Default);
	virtual ~GObject() {}

	// set/get
	virtual ClassType isA() const = 0;
	virtual const char* typeStr() const = 0;

	void	name(const std::string& aname)	{ _name = aname; }
const	std::string& name()		const	{ return _name; }
const	char*	cname()			const	{ return _name.c_str(); }
	void	id(const int ii)		{ _id = ii; }
	int	id()			const	{ return _id; }

	virtual	void position(const Point& r)	{ P = r; }
	Point	position()		const	{ return P; }
	Point	savedPosition()		const	{ return SP; }
	Point	viewerPosition()	const	{ return VP; }

	int	config(PyObject*);
	virtual PyObject* config(const char* key, PyObject* value=NULL);

	bool	visible() const { return _visible && Abs(VP.z)<=drawDw; }

	// save&recall
	virtual	void	save()			{ SP = P; }
	virtual void	restore()		{ P  = SP; }
	virtual Point	savedNode(int)	const	{ return SP; }

	// transform
	virtual int	nodes() const = 0;
	virtual Point	node(int)	const	{ return P; }
	virtual	void	node(int item, const Point& r);
	virtual	void	rotate(const double, const Vector& /*axis*/) {}
	virtual void	transform(ViewerObject* self);

	// bounding box
	virtual BBox	bbox() const;
	virtual BBox	bboxView(ViewerObject* self);

	// drawing routines
	virtual void	draw(ViewerObject* self, Drawable drawable);
	virtual void	drawText(ViewerObject* self, Drawable drawable);

	// closer routines
	virtual int	closest(ViewerObject* self, const int ii, const int jj, const int d);
	virtual bool	enclosed(ViewerObject* self, const int l, const int t, const int r, const int b);

protected:
	int	close2Line(ViewerObject* self, const int ii, const int jj, const int d,
			const Point& A, const Point B) const;
	int	close2Point(const int ii, const int jj, const int xx, const int yy, const int d) const {
			return Sq((double)(ii-xx)) + Sq((double)(jj-yy)) <= (double)Sq(size+d);
			//return Sq(ii-xx) + Sq(jj-yy) <= Sq(size+d);
		}
	void	setForeground(ViewerObject* self) const;
	void	drawPoint(ViewerObject* self, Drawable drawable, int ii, int jj, int s) const;
	void	drawSelectedPoint(ViewerObject* self, Drawable drawable, int ii, int jj, int s) const;
}; // GObject

// ================================== GPoint ===================================
class GPoint : public GObject {
public:
	GPoint(const char* aname=NULL, ObjectType o=Point_Cross) : GObject(aname, o) {};
	virtual ~GPoint() {}

	virtual ClassType isA()		const	{ return GPointClass; }
	virtual const char* typeStr()	const	{ return "point"; }

	// nodes
	virtual int nodes() const { return 1; }

	virtual void draw(ViewerObject* self, Drawable drawable);
}; // GPoint

// ================================== GArrow ===================================
class GArrow : public GObject {
public:
	enum NodeId : int {
		NodeId_Head = 1,
		NodeId_Tail = 2,
	};
protected:
	Vector	D;			// vector length
	Vector	SD;			// save variables

	Point	Ve, Vec;		// Viewer end position (clipped)
	int	x2, y2;

	bool	drawHead;		// draw a head arrow
	bool	drawTail;		// draw a tail arrow

public:
	GArrow(const char* aname=NULL, ObjectType o=Arrow_Head)
		: GObject(aname, o), D(10.,10.,10.), x2(0), y2(0), drawHead(true), drawTail(false) {}
	virtual ~GArrow() {}

	virtual ClassType isA()		const	{ return GArrowClass; }
	virtual const char* typeStr()	const	{ return "arrow"; }
	virtual	PyObject* config(const char* key, PyObject* value=NULL);

		void	direction(const double x, const double y, const double z) { D.x=x; D.y=y; D.z=z; }
		void	direction(const Vector& d)	{ D = d; }
		Vector	direction()	const	{ return D; }
		Point	viewerEnd()	const	{ return Ve; }

	// save & recall
	virtual	void	save();
	virtual void	restore();
	virtual Point	savedNode(int)	const;

	// transform
	virtual	int	nodes()		const	{ return 2; }
	virtual Point	node(int)	const;
	virtual	void	node(int item, const Point& r);
	virtual	void	rotate(const double angle, const Vector& axis);
	virtual void	transform(ViewerObject* self);

	// bounding box
	virtual BBox	bbox() const;
	virtual BBox	bboxView(ViewerObject* self);

	// drawing routines
	virtual void	draw(ViewerObject* self, Drawable drawable);
	virtual void	drawText(ViewerObject* self, Drawable drawable);

	// closer routines
	virtual int	closest(ViewerObject* self, const int ii, const int jj, const int d);
	virtual bool	enclosed(ViewerObject* self, const int left, const int top,
				const int right, const int bottom);
}; // GArrow

// ================================== GSpline ==================================
class GSpline : public GObject {
protected:
	CardinalSpline<VectorSplineNode> spline;// time spline information
	std::vector<double>	nodetime;	// node time used for integration
	std::vector<Point>	Vpoint;		// viewer points
	std::vector<Vector>	Spoint;		// saved node points
	int	duplicates;			// duplicate nodes. Do not use them in the spline
	int	ndrawsegs;			// actual drawing segments (from transform)
						// only in time
public:
	int	nsegments;			// drawing segments requested

public:
	GSpline(const char* aname=NULL, ObjectType o=Point_Cross)
			: GObject(aname, o),
			  duplicates(0),
			  nsegments(3)
			  {}
	virtual ~GSpline() {}

	virtual ClassType isA()		const	{ return GSplineClass; }
	virtual const char* typeStr()	const	{ return "spline"; }
	virtual	PyObject* config(const char* key, PyObject* value=NULL);

	// get
	CardinalSpline<VectorSplineNode>& getSpline()	{ return spline; }
	double	getTime(int i)	const	{ return InRange(0,i,(int)nodetime.size()-1)? nodetime[i] : 0.0; }
	double	getTension()	const	{ return spline.tension(); }

	// save&recall
	virtual	void	save();
	virtual void	restore();
	virtual Point	savedNode(int) const;

	// transform
	virtual int	nodes()		const	{ return spline.size(); }
	virtual Point	node(int)	const;
	virtual	void	node(int item, const Point& r);
	//virtual	void	rotate(const double, const Vector& axis) {}
	virtual void	transform(ViewerObject* self);

	// bounding box
	virtual BBox	bbox() const;
	virtual BBox	bboxView(ViewerObject* self);

	// drawing routines
	virtual void	draw(ViewerObject* self, Drawable drawable);

	// closer routines
	virtual int	closest(ViewerObject* self, const int ii, const int jj, const int d);

private:
	void	scanDuplicates();
}; // GSpline

// ================================== GRuler ===================================
class GRuler : public GArrow {
public:
	enum NodeId : int {
		NodeId_Angle = 3,
	};
protected:
	Vector	Da;		// angle point
	Vector	SDa;		// save

	Point	Va, Vac;	// viewer angle position + clipped
	Point	Vc2;		// viewer angle-starting position clipped

	int	x3, y3;

public:
	GRuler(const char* aname=NULL, ObjectType o=Ruler_Simple)
		: GArrow(aname,o), Da(10.,0.,0.) {}
	virtual ~GRuler() {}

	virtual ClassType isA()		const	{ return GRulerClass; }
	virtual const char* typeStr()	const	{ return "ruler"; }

	virtual PyObject* config(const char* key, PyObject* value=NULL);

		Vector	viewerAngle()	const	{ return Va; }

	// save & recall
	virtual	void	save();
	virtual void	restore();
	virtual Point	savedNode(int)	const;

	// transform
	virtual int	nodes()		const	{ return (type==Ruler_Angle?3:2); }
	virtual Point	node(int)	const;
	virtual	void	node(int item, const Point& r);
	virtual	void	rotate(const double angle, const Vector& axis);
	virtual void	transform(ViewerObject* self);

	// bounding box
	virtual BBox	bbox() const;
	virtual BBox	bboxView(ViewerObject* self);

	// drawing routines
	virtual void	draw(ViewerObject* self, Drawable drawable);
	virtual void	drawText(ViewerObject* self, Drawable drawable);

	// closer routines
	virtual int	closest(ViewerObject* self, const int ii, const int jj, const int d);
	virtual bool	enclosed(ViewerObject* self, const int left, const int top,
				const int right, const int bottom);
}; // GRuler

// ================================= GRotdefi ==================================
class GRotdefi : public GArrow {
public:
	Matrix4	orient;			// Initial orientation/position of base
	Matrix4	matrix;			// Rotdefi matrix for projection
	int	axisSize;		// axis size
	int	axisWidth;		// axis line thickness

public:
	GRotdefi(const char* aname=NULL) : GArrow(aname,Arrow_Head),
			orient(1),
			matrix(1),
			axisSize(20),
			axisWidth(3)
			{ }
	virtual ~GRotdefi() {}

	virtual ClassType isA()		const	{ return GRotdefiClass; }
	virtual const char* typeStr()	const	{ return "axis"; }

	virtual PyObject* config(const char* key, PyObject* value=NULL);

	// nodes
//	virtual int nodes() const { return 4; }

	// bounding box
//	virtual BBox	bbox() const;
//	virtual BBox	bboxView(ViewerObject* self);

	// drawing routines
	virtual void	draw(ViewerObject* self, Drawable drawable);
//	virtual void	drawText(ViewerObject* self, Drawable drawable);

	// closer routines
	virtual int	closest(ViewerObject* self, const int ii, const int jj, const int d);
	virtual bool	enclosed(ViewerObject* self, const int left, const int top,
				const int right, const int bottom);
}; // GRotdefi

// ================================== GLight ===================================
class GLight : public GArrow {
public:
	enum NodeId : int {
		NodeId_Radius = 2,
	};
public:
	double	power, spec;		// light power and specular
	bool	relative;		// relative to origin or absolute
	bool	shadow;			// shadow casting
	Light::FallOff	falloff;	// falloff type 0,1,2
	int	nsamples;		// number of ray samples
	double	radius;			// volumetric size (radius) of light

public:
	GLight(const char* aname=NULL, ObjectType o=Light_Sun);
	virtual ~GLight() {}

	virtual ClassType isA()		const	{ return GLightClass; }
	virtual const char* typeStr()	const	{ return "light"; }

	virtual PyObject* config(const char* key, PyObject* value=NULL);

	// transform
	virtual int	nodes()		const	{ return 2; }
	virtual Point	node(int n)	const	{ return GArrow::node(n); }
	virtual	void	node(int item, const Point& r);
	//virtual void	transform(ViewerObject* self);

	virtual void	draw(ViewerObject* self, Drawable drawable);
	virtual void	drawText(ViewerObject* self, Drawable drawable);

	// closer routines
	virtual int	closest(ViewerObject* self, const int ii, const int jj, const int d);

	// convert to light structure
	void toLight(Light& L) const;
}; // GLight

// =================================== GBeam ===================================
class GBeam : public GArrow {
public:
	double	energy;		// beam energy
	double	scale;		// scaling of arrow
	double	divergence2;	// beam half divergence
	double	Rin, Rout;	// ring radius

public:
	GBeam(const char* aname=NULL, ObjectType o=Beam_Arrow);
	virtual ~GBeam() {}

	virtual ClassType isA()		const	{ return GBeamClass; }
	virtual const char* typeStr()	const	{ return "beam"; }

	virtual PyObject* config(const char* key, PyObject* value=NULL);

	// transform
//	virtual int	nodes()		const	{ return 2; }
//	virtual Vector	node(int n)	const	{ return GArrow::node(n); }
//	virtual	void	node(int item, const Point& r);
//	//virtual void	transform(ViewerObject* self);

	virtual void	draw(ViewerObject* self, Drawable drawable);
//	virtual void	drawText(ViewerObject* self, Drawable drawable);

	// closer routines
//	virtual int	closest(ViewerObject* self, const int ii, const int jj, const int d);
}; // GBeam

// ================================== GCamera ==================================
class GCamera : public GObject {
public:
	enum NodeId : int {
		NodeId_At    = 2,
		NodeId_Focal = 3,
		NodeId_Up    = 4,
	};
protected:
	Point	at;		// look at point
	Vector	up;		// up pointing vector
	double	focal;		// focal length
	double	fov;		// field of view (2*aperture angle)
	double	aspect;		// aspect ratio Width/Height
private:
	double	w2,h2;		// half width horizontal and vertical
	Vector	X,Y,Z;		// X=horizontal, Y=vertical, Z=camera look at direction
	Point	Vat, Vatc;	// Viewer at
	Vector	VX, VY, VZ;	// Viewer X&Y

	Point	Vpoint[5];	// drawing points

	Point	Sat;		// saved at
	Vector	Sup;		// saved up
	double	Sfocal;		// saved focal
	Vector	SY, SZ;		// saved Y&Z

public:
	GCamera(const char* aname=NULL, ObjectType o=Camera_Persp)
		: GObject(aname, o),
		  focal(10.),
		  fov(Rad(90.)),
		  aspect(4./3.)
		{}
	virtual ~GCamera() {}

	virtual ClassType isA()		const	{ return GCameraClass; }
	virtual const char* typeStr()	const	{ return "camera"; }

	virtual PyObject* config(const char* key, PyObject* value=NULL);

	virtual	void position(const Point& r) {
			GObject::position(r);
			init();
		}

	// initialize vectors
		void	init();
	virtual	void	save();
	virtual void	restore();
	virtual Point	savedNode(int) const;

	// transform
	virtual int	nodes()	const		{ return 4; }
	virtual Point	node(int n) const;
	virtual	void	node(int item, const Point& r);
	virtual void	transform(ViewerObject* self);

	// drawing routines
	virtual void	draw(ViewerObject* self, Drawable drawable);

	// closer routines
	virtual int	closest(ViewerObject* self, const int ii, const int jj, const int d);
}; // class GCamera

#endif
