/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	16-Jul-2010
 */

#include "painter.h"

#include <assert.h>
#include <string.h>
#include <stdarg.h>

#include "math/bit.h"

// if wchar_t is 4 bytes use the wmemset for faster background setting
#if WCHAR_MAX > 0xffffu
	//wchar_t is 4 bytes
#	define WCHAR4
#endif

const int MinSize = 2;
const int MaxSize = 1<<15;

/* ============================= Painter ================================= */
/** Painter */
Painter::Painter() : //const int w, const int h) :
	_width(0),
	_height(0),
	_dataSize(0),
	_maxSize(0),
	_dpiscale(0),
	_data(NULL)
{
	background(0x707070);
} // Painter

/** ~Painter */
Painter::~Painter()
{
	if (_data) delete[] _data;
} /* ~Painter */

/** initialize buffer dimensions
 * @param w	width of image
 * @param h	height of image
 * @return true if a new buffer is allocated, false if the existing is used instead
 */
bool Painter::init(const int w, const int h)
{
	_width  = Range(MinSize, w, MaxSize);
	_height = Range(MinSize, h, MaxSize);
	_clip.set(_width-1, _height-1);

	_dataSize = _width*_height;
	if (_dataSize > _maxSize) {
		if (_data)	delete[] _data;
		_maxSize = _dataSize;
		if (_maxSize < 1<<23 && !isPowerOf2(_maxSize))
			_maxSize = roundUpPow2(_maxSize);
		_data = new dword[_maxSize];
		return true;
	} else
		return false;
} // init

/** clear image */
void Painter::clear()
{
	if (_clip.isclipped(_width-1, _height-1)) {
		int mx = _clip.width();
		int my = _clip.height();
		dword* src = pixelPtr(_clip.left, _clip.top);
		for (int i=0; i<my; i++) {
			std::fill_n(src, mx, _background);
			src += _width;
		}
	} else
		std::fill_n(_data, _dataSize, _background);
} // clear

/** move image by (dx,dy)
 * @param dx,dy	pixels to move image
 */
void Painter::move(const int dx, const int dy)
{
	dword *src, *dst;
	int mx = (_width - Abs(dx))*(int)sizeof(dword);
	int my = _height - Abs(dy);

	if (dy >= 0) {
		if (dx>=0) {
			src = pixelPtr( 0, _height-1-dy);
			dst = pixelPtr(dx, _height-1);
		} else {	/* dx<0 */
			src = pixelPtr(-dx, _height-1-dy);
			dst = pixelPtr(  0, _height-1);
		}
		for (int j=0; j<my; j++) {
			memmove(dst, src, mx);
			dst -= _width;
			src -= _width;
		}
	} else {
		if (dx>=0) {
			src = pixelPtr( 0,-dy);
			dst = pixelPtr(dx,  0);
		} else {	/* dx<0 */
			src = pixelPtr(-dx,-dy);
			dst = pixelPtr(  0,  0);
		}
		for (int j=0; j<my; j++) {
			memmove(dst, src, mx);
			dst += _width;
			src += _width;
		}
	}
} // move

/** draw pixel at (x,y) with color using intensity from 0..1
 * @param x,y		position to draw
 * @param color		color to use
 * @param intensity	intensity of the color*intensity
 */
void Painter::pixel(const int x, const int y, dword color, double intensity)
{
	Color32 c(color);
	Color32 p(pixel(x,y));

	c.red()   = (int)((double)c.red()   * intensity + (double)p.red()   * (1.0-intensity));
	c.green() = (int)((double)c.green() * intensity + (double)p.green() * (1.0-intensity));
	c.blue()  = (int)((double)c.blue()  * intensity + (double)p.blue()  * (1.0-intensity));
	pixel(x,y,c);
} // pixel

/** unclippedLine
 * @param x1,y1	starting point of line
 * @param x2,y2	end point of line
 * @param color	color of line
 */
void Painter::unclippedLine(int x1, int y1, int x2, int y2, const dword color)
{
	int dx = Abs(x1-x2);
	int dy = Abs(y1-y2);

	if (dx > dy) {
		if (x1 > x2) {			// force x1<x2
			Swap(x1,x2);
			Swap(y1,y2);
		}
		int yinc = (y2>y1)? 1: -1;	// determine increment for y
		int d = 2*dy - dx;
		int Ainc = 2*(dy-dx);
		int Binc = 2*dy;

		int x = x1;
		int y = y1;
		pixel(x,y,color);

		for (x=x1+1; x<=x2; x++) {
			if (d >= 0) {		// setpixel A
				y += yinc;
				d += Ainc;
			} else
				d += Binc;	// setpixel B
			pixel(x,y,color);
		}
	} else {				/* DO THE SAME FOR Y */
		if (y1 > y2) {			// force y1<y2
			Swap(x1,x2);
			Swap(y1,y2);
		}
		int xinc = (x2>x1)? 1: -1;	// determine increment for x
		int d = 2*dx - dy;
		int Ainc = 2*(dx-dy);
		int Binc = 2*dx;

		int x = x1;
		int y = y1;
		pixel(x,y,color);

		for (y=y1+1; y<=y2; y++) {
			if (d >= 0) {		// setpixel A
				x += xinc;
				d += Ainc;
			} else
				d += Binc;	// setpixel B
			pixel(x,y,color);
		}
	}
} // unclippedLine

/** unclippedThickLine
 * @param x1,y1	starting point of line
 * @param x2,y2	end point of line
 * @param w	line thickness in pixels
 * @param color	color of line
 */
void Painter::unclippedThickLine(int x1, int y1, int x2, int y2, const int w, const dword color)
{
	int   d, dx, dy;
	int   Ainc, Binc;
	int   x, y;

	dx = Abs(x1-x2);
	dy = Abs(y1-y2);

	if (!dx && !dy) {
		// fill a square
		int w2 = w/2;
		for (y=_clip.cropTop(y1-w2); y<=_clip.cropBottom(y1+w2); y++)
			for (x=_clip.cropLeft(x1-w2); x<=_clip.cropRight(x1+w2); x++)
				pixel(x,y,color);
		return;
	}

	if (dx > dy) {
		if (x1 > x2) {			/* force x1<x2 */
			Swap(x1,x2);
			Swap(y1,y2);
		}
		int l = (int)(w * sqrt(Sq(dx)+Sq(dy)) / dx / 2);
		if (l<1) l=1;
		int yinc = (y2>y1)? 1: -1;	/* determine increment for y */
		d = 2*dy - dx;
		Ainc = 2*(dy-dx);
		Binc = 2*dy;

		x = x1;
		y = y1;
		for (int yy=_clip.cropTop(y-l); yy<_clip.cropBottom(y+l); yy++)
			pixel(x,yy,color);

		for (x=x1+1; x<=x2; x++) {
			if (d >= 0) {		/* setpixel A */
				y += yinc;
				d += Ainc;
			} else
				d += Binc;	/* setpixel B */
			for (int yy=_clip.cropTop(y-l); yy<_clip.cropBottom(y+l); yy++)
				pixel(x,yy,color);
		}
	} else {	/* DO THE SAME BUT FOR Y */
		if (y1 > y2) {			/* force y1<y2 */
			Swap(y1,y2);
			Swap(x1,x2);
		}
		int l = (int)(w * sqrt(Sq(dx)+Sq(dy)) / dy / 2);
		if (l<1) l=1;
		int xinc = (x2>x1)? 1: -1;	/* determine increment for x */
		d = 2*dx - dy;
		Ainc = 2*(dx-dy);
		Binc = 2*dx;

		x = x1;
		y = y1;
		for (int xx=x-l; xx<x+l; xx++)
			pixel(xx,y,color);

		for (y=y1+1; y<=y2; y++) {
			if (d >= 0) {		/* setpixel A */
				x += xinc;
				d += Ainc;
			} else
				d += Binc;	/* setpixel B */
			for (int xx=_clip.cropLeft(x-l); xx<_clip.cropRight(x+l); xx++)
				pixel(xx,y,color);
		}
	}
} // unclippedThickLine

/** Drawing of anti-aliased lines following Wu's algorithm, based on M. Abrash'
 * excellent graphics programming black book. One can still improve the AA
 * quality & a number of points, but for the moment this method gives quite
 * satisfying results with reasonable speed.
 *
 * @param x1,y1	starting point of line
 * @param x2,y2	end point of line
 * @param color	color of line
 * @param alpha	if true set the color and adjust the alpha channel
 *
 * [theis] 13/1/2020
 */
void Painter::unclippedLineAntialias(int x1, int y1, int x2, int y2, dword lineColor, bool alpha)
{
	Color32 color(lineColor);

	// ensure orientation (top -> bottom)
	if (y1 > y2) {
		Swap(y1, y2);
		Swap(x1, x2);
	}

	// start pixel
	pixel(x1, y1, lineColor);

	// find orientation in X
	int deltaX = x2-x1;
	int xdir = x2>x1? 1 : -1;
	if (xdir<0) deltaX = -deltaX;

	// find orientation in Y
	int deltaY = y2-y1;

	// horizontal & vertical lines need no anti-aliasing
	if (deltaY==0) {
		while (deltaX-- != 0) {
			x1 += xdir;
			pixel(x1, y1, lineColor);
		}
		return;
	}
	if (deltaX==0) {
		do {
			y1++;
			pixel(x1, y1, lineColor);
		} while (--deltaY != 0);
		return;
	}
	// diagonal lines need it neither
	// FIXME XXX Questionable!!!
	if (deltaX==deltaY) {
		do {
			x1 += xdir;
			y1++;
			pixel(x1, y1, lineColor);
		} while (--deltaY != 0);
		return;
	}

	// treat tilted lines
	word errorAdj;
	word tempAccumulatedError, weight, weightComplement;
	word errorAccumulation = 0;

	byte rl = color.R();
	byte gl = color.G();
	byte bl = color.B();
	float grayl = color.gray();

	//
	// Y-dominant lines
	//
	if (deltaY > deltaX) {
		// determine the fractional part that X advances each time Y
		// advances. We store this in 0.16 format & whenever this turns
		// over, we have to advance by one pixel in the minor axis
		// Results are truncated to avoid overruns.
		errorAdj = ((unsigned long)deltaX << 16) / (unsigned long)deltaY;

		while (--deltaY) {
			tempAccumulatedError = errorAccumulation;
			errorAccumulation += errorAdj;      // calculate error for next pixel

			// in case of a turn-over we have to advance the X coordinate as well.
			// Y is always advanced
			if (errorAccumulation <= tempAccumulatedError)
				x1 += xdir;
			y1++;

			// the weight and its complement for the paired pixel
			// is determined by the distance from the pixel center
			// (= accumulated error) It is given by the 8 most
			// significant bits.
			weight = errorAccumulation >> 8;
			weightComplement = weight ^ 255;
			assert(weight < 256);
			assert(weightComplement < 256);

			if (alpha) {
				color.alpha() = (FLAG_ALPHA>>24) | (weightComplement>>1);
				pixel(x1, y1, color);
				color.alpha() = (FLAG_ALPHA>>24) | (weight>>1);
				pixel(x1+xdir, y1, color);
			} else {
				Color32 curCol(pixel(x1, y1));
				byte rb = curCol.R();
				byte gb = curCol.G();
				byte bb = curCol.B();
				float grayb = curCol.gray();

				// calculate blended color considering the weight. The
				// further away from the actual pixel center the lower
				// the weight This is basically a simple box filter
				byte r = rb>rl? ((byte)(((float)(grayl<grayb? weight : weightComplement)) / 255.0 * (rb-rl) + rl))
					      : ((byte)(((float)(grayl<grayb? weight : weightComplement)) / 255.0 * (rl-rb) + rb));
				byte g = gb>gl? ((byte)(((float)(grayl<grayb? weight : weightComplement)) / 255.0 * (gb-gl) + gl))
					      : ((byte)(((float)(grayl<grayb? weight : weightComplement)) / 255.0 * (gl-gb) + gb));
				byte b = bb>bl? ((byte)(((float)(grayl<grayb? weight : weightComplement)) / 255.0 * (bb-bl) + bl))
					      : ((byte)(((float)(grayl<grayb? weight : weightComplement)) / 255.0 * (bl-bb) + bb));
				pixel(x1, y1, Color32::RGB(r,g,b));

				Color32 nextPixelCol(pixel(x1+xdir, y1));
				rb = nextPixelCol.R();
				gb = nextPixelCol.G();
				bb = nextPixelCol.B();
				grayb = nextPixelCol.gray();

				r = rb>rl? ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (rb-rl) + rl))
					 : ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (rl-rb) + rb));
				g = gb>gl? ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (gb-gl) + gl))
					 : ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (gl-gb) + gb));
				b = bb>bl? ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (bb-bl) + bl))
					 : ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (bl-bb) + bb));
				pixel(x1+xdir, y1, Color32::RGB(r,g,b));
			}
		}

		// Draw the final pixel, which is always exactly intersected by
		// the line and so needs no weighting
		pixel(x2, y2, lineColor);
		return;
	}

	//
	// X dominant line
	//
	// determine the fractional part that Y advances each time X advances.
	// We store this in 0.16 format & whenever this turns over, we have to
	// advance by one pixel in the minor axis Results are truncated to
	// avoid overruns.
	errorAdj = ((unsigned long)deltaY << 16) / (unsigned long)deltaX;

	while (--deltaX) {
		tempAccumulatedError = errorAccumulation;
		errorAccumulation += errorAdj;			// calculate error for next pixel

		// in case of a turn-over we have to advance the Y coordinate
		// as well. X is always advanced
		if (errorAccumulation <= tempAccumulatedError)
			y1++;
		x1 += xdir;

		// the weight and its complement for the paired pixel is
		// determined by the distance from the pixel center (=
		// accumulated error) It is given by the 8 most significant
		// bits.
		weight = errorAccumulation >> 8;
		weightComplement = weight ^ 255;
		assert(weight < 256);
		assert(weightComplement < 256);

		if (alpha) {
			color.alpha() = (FLAG_ALPHA>>24) | (weightComplement>>1);
			pixel(x1, y1, color);

			color.alpha() = (FLAG_ALPHA>>24) | (weight>>1);
			pixel(x1, y1+1, color);
		} else {
			Color32 curCol(pixel(x1, y1));
			byte rb = curCol.R();
			byte gb = curCol.G();
			byte bb = curCol.B();
			float grayb = curCol.gray();

			// calculate blended color considering the weight. The further
			// away from the actual pixel center the lower the weight This
			// is basically a simple box filter
			byte r = rb > rl ? ((byte)(((float)(grayl < grayb? weight : weightComplement)) / 255.0 * (rb - rl) + rl))
					 : ((byte)(((float)(grayl < grayb? weight : weightComplement)) / 255.0 * (rl - rb) + rb));
			byte g = gb > gl ? ((byte)(((float)(grayl < grayb? weight : weightComplement)) / 255.0 * (gb - gl) + gl))
					 : ((byte)(((float)(grayl < grayb? weight : weightComplement)) / 255.0 * (gl - gb) + gb));
			byte b = bb > bl ? ((byte)(((float)(grayl < grayb? weight : weightComplement)) / 255.0 * (bb - bl) + bl))
					 : ((byte)(((float)(grayl < grayb? weight : weightComplement)) / 255.0 * (bl - bb) + bb));
			pixel(x1, y1, Color32::RGB(r,g,b));

			Color32 nextPixelCol(pixel(x1, y1+1));
			rb = nextPixelCol.R();
			gb = nextPixelCol.G();
			bb = nextPixelCol.B();
			grayb = nextPixelCol.gray();

			r = rb>rl? ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (rb-rl) + rl))
				 : ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (rl-rb) + rb));
			g = gb>gl? ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (gb-gl) + gl))
				 : ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (gl-gb) + gb));
			b = bb>bl? ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (bb-bl) + bl))
				 : ((byte)(((float)(grayl<grayb? weightComplement : weight)) / 255.0 * (bl-bb) + bb));
			pixel(x1, y1+1, Color32::RGB(r,g,b));
		}
	}

	// draw the final pixel
	pixel(x2, y2, lineColor);
} // unclippedLineAntialias

/** unclippedBeautyPass
 * draw a line and add the ALPHA flag to perform anti-aliasing as a post-draw effect
 * based on pixels flagged to be treated in the beauty pass
 *
 * @param x1,y1	starting point of line
 * @param x2,y2	end point of line
 * @param color	color of line
 *
 * [theis] 20/1/2020
 */
void Painter::unclippedLineBeautyPass(int x1, int y1, int x2, int y2, dword lineColor)
{
	if (x2 != x1 && y2 != y1)  // apply to non-horizontal and non-vertical lines only
		lineColor |= FLAG_ALPHA;
	unclippedLine(x1, y1, x2, y2, lineColor);
} // unclippedLineBeautyPass

/** draw rectangle
 * @param x1,y1	upper-left corner
 * @param x2,y2 lower-right corner
 * @param color	color to draw
 */
bool Painter::rectangle(int x1, int y1, int x2, int y2, const dword color)
{
	bool b = line(x1,y1, x2,y1, color);
	b |= line(x2,y1, x2,y2, color);
	b |= line(x2,y2, x1,y2, color);
	b |= line(x1,y2, x1,y1, color);
	return b;
} // rectangle

/** fill rectangle
 * @param x1,y1	upper-left corner
 * @param x2,y2 lower-right corner
 * @param color	color to draw
 */
void Painter::fillRect(int x1, int y1, int x2, int y2, const dword color)
{
	_clip.point(&x1, &y1);
	_clip.point(&x2, &y2);

	dword *src = pixelPtr(x1,y1);
	for (int i=x1; i<=x2; i++)
		*src++ = color;

	int w = x2-x1+1;
	src -= w;
	dword *dst = src + _width;
	w *= (int)sizeof(*_data);
	for (int j=y1+1; j<=y2; j++) {
		memcpy(dst, src, w);
		dst += _width;
	}
} // fillRect

/** lighten or darken rectangle
 * @param x1,y1	upper-left corner
 * @param x2,y2 lower-right corner
 * @param level	to shift colors
 */
void Painter::levelShiftRect(int x1, int y1, int x2, int y2, int level)
{
	if (level==0) return;

	_clip.point(&x1, &y1);
	_clip.point(&x2, &y2);

	if (level > 0) {
		for (int j=y1; j<=y2; j++) {
			dword *ptr = pixelPtr(x1,j);
			for (int i=x1; i<=x2; i++, ptr++)
				*ptr = Color32::lighter(*ptr, level);
		}
	} else {
		level = -level;
		for (int j=y1; j<=y2; j++) {
			dword *ptr = pixelPtr(x1,j);
			for (int i=x1; i<=x2; i++, ptr++)
				*ptr = Color32::darker(*ptr, level);
		}
	}
} // levelShiftRect

/** draw bitmap
 * @param x,y	upper-left corner where to draw
 * @param w,h	width and height of the bitmap
 * @param bitmap	bitmap data of size w*h
 * @return	true if it draw something, else false
 */
bool Painter::drawBitmap(int x, int y, int w, int h, const dword* bitmap)
{
	int xs=x,     ys=y;
	int xe=x+w-1, ye=y+h-1;

	// Clip starting and ending point
	//
	_clip.point(&xs, &ys);
	_clip.point(&xe, &ye);

	if (xe-xs<1 || ye-ys<1) return false;

	const Color32* src = (const Color32*)bitmap;
	dword* dst   = pixelPtr(xs,ys);

	// skip pixels outside
	src += (ys-y)*w + (xs-x);

	int dw = xe-xs+1;
	w -= dw;
	int Dw = _width-dw;

	//bool found = false;
	for (int j=ys; j<=ye; j++) {
		for (int i=xs; i<=xe; i++) {
			if (src->alpha() != 0xFF)
				*dst++ = src->value;
			else
				dst++;
			src++;
		}
		src +=  w;
		dst += Dw;
	}
	return true;
} // drawBitmap

/** flood fill similar colors
 * @param x,y	starting point
 * @param color	color to fill
 */
void Painter::fill(int x, int y, const dword color, const dword color2, const Fill filltype)
{
#define STACKSIZE 500
	int stackX[STACKSIZE];
	int stackY[STACKSIZE];

	dword *pixelup, *pixeldown;

	if (!_clip.inside(x,y)) return;
	dword startColor = pixel(x,y);
	if (startColor==color or (filltype!=Fill::FLOOD and startColor==color2)) return;

	int tos = 0;					// top of stack
	for (;;) {					// do forever
		while (pixel(x,y)!=startColor) {	// pick an empty point
			if (!tos) return;
			x = stackX[--tos];
			y = stackY[  tos];
		}

		int xrem = x+1;
		dword* pix = pixelPtr(x,y);		// get pointer

		/* scan line */
		if (_clip.top<y)
			pixelup = pix - _width;		// up one line
		else
			pixelup = NULL;

		if (y<_clip.bottom)
			pixeldown = pix + _width;	// down one line
		else
			pixeldown = NULL;

		int startup   = (pixelup && *pixelup==startColor);
		int startdown = (pixeldown && *pixeldown==startColor);

		int up   = 0;
		int down = 0;
		while (*pix == startColor) {
			switch (filltype) {
				case Fill::FLOOD:
					*pix = color;
					break;
				case Fill::DOTS:
					*pix = (((x+y)&3) || (y&1))? color : color2;
					break;
				case Fill::HASH:
					*pix = (x+y)&7? color : color2;
					break;
				case Fill::HASHR:
					*pix = (x-y)&7? color : color2;
					break;
				case Fill::X:
					*pix = (((x+y)&3) && (x-y)&3)? color : color2;
					break;
				case Fill::X2:
					*pix = (((x+y)&7) && (x-y)&7)? color : color2;
					break;
				default:
					assert(0);
			}

			int pup = (pixelup && *pixelup==startColor);
			if (!up && pup && tos<STACKSIZE) {
				stackX[tos  ] = x;
				stackY[tos++] = y-1;
			}
			up = pup;

			int pdown = (pixeldown && *pixeldown==startColor);
			if (!down && pdown && tos<STACKSIZE) {
				stackX[tos  ] = x;
				stackY[tos++] = y+1;
			}
			down = pdown;
			if (x==_clip.left) break;
			x--;
			pix--;
			if (pixelup) pixelup--;
			if (pixeldown) pixeldown--;
		}

		if (xrem>_clip.right) continue;
		x = xrem;

		pix = pixelPtr(x,y);

		/* scan line */
		if (_clip.top<y)
			pixelup = pix - _width;		// up one line
		else
			pixelup = NULL;

		if (y<_clip.bottom)
			pixeldown = pix + _width;	// down one line
		else
			pixeldown = NULL;

		up   = startup;
		down = startdown;
		while (*pix == startColor) {
			switch (filltype) {
				case Fill::FLOOD:
					*pix = color;
					break;
				case Fill::DOTS:
					*pix = (((x+y)&3) || (y&1))? color : color2;
					break;
				case Fill::HASH:
					*pix = (x+y)&7? color : color2;
					break;
				case Fill::HASHR:
					*pix = (x-y)&7? color : color2;
					break;
				case Fill::X:
					*pix = (((x+y)&3) && (x-y)&3)? color : color2;
					break;
				case Fill::X2:
					*pix = (((x+y)&7) && (x-y)&7)? color : color2;
					break;
				default:
					assert(0);
			}

			int pup = (pixelup && *pixelup==startColor);
			if (!up && pup && tos<STACKSIZE) {
				stackX[tos  ] = x;
				stackY[tos++] = y-1;
			}
			up = pup;

			int pdown = (pixeldown && *pixeldown==startColor);
			if (!down && pdown && tos<STACKSIZE) {
				stackX[tos  ] = x;
				stackY[tos++] = y+1;
			}
			down = pdown;
			if (x==_clip.right) break;
			x++;
			pix++;
			if (pixelup) pixelup++;
			if (pixeldown) pixeldown++;
		}
	}
} // fill

/** copy */
void Painter::copy(Painter& src)
{
	assert(width() == src.width());
	assert(height() == src.height());
	memcpy(_data, src._data, _dataSize);
} // copy

/** fit source image to current painter, by rescaling */
void Painter::fit(Painter& src, Painter::Fit fittype)
{
	if (width() == src.width() and
	    height() == src.height()) {
		copy(src);
		return;
	}
	//float aspect  = (float)width() / (float)height();
	//float daspect = (float)src.width() / (float)src.height();
	//if (aspect > daspect) {

	std::fill_n(_data, _dataSize, _background);

	int i,j, si,sj;
	dword* pix = _data;
	switch (fittype) {
		case Fit::SCALE:
			for (j=0; j<_height; j++) {
				sj = (j * src.height()) / height();
				for (i=0; i<_width; i++) {
					si = (i * src.width()) / width();
					*pix++ = src.pixel(si,sj);
				}
			}
			break;
		// case FIXME other interpolation methods
		default:
			assert(false);
	}
} // fit

/** measure string */
int Painter::measuref(BFont& font, const char* fmt, ...)
{
	char	text[1024];
	va_list	ap;
	if (fmt==NULL) return 0;
	va_start(ap, fmt);
	vsprintf(text, fmt, ap);
	va_end(ap);

	return measure(font, text)*dpiscale();
} // measuref

/** print formatted string
 * @return right most pixel of the string
 */
int Painter::printf(BFont& font, int x, int y, const dword color, const char* fmt, ...)
{
	char	text[1024];
	va_list	ap;
	if (fmt==NULL) return x;
	va_start(ap, fmt);
	vsprintf(text, fmt, ap);
	va_end(ap);

	for (byte *ch=(byte*)text; *ch; ch++)
		if (_dpiscale)
			x += font.drawScaled(*this, x, y, color, *ch, _dpiscale);
		else
			x += font.draw(*this, x, y, color, *ch);

	return x;
} // printf

/** vertical print formatted string
 * @return right most pixel of the string
 */
int Painter::vprintf(BFont& font, int x, int y, const dword color, const char* fmt, ...)
{
	char	text[1024];
	va_list	ap;
	if (fmt==NULL) return x;
	va_start(ap, fmt);
	vsprintf(text, fmt, ap);
	va_end(ap);

	for (byte *ch=(byte*)text; *ch; ch++)
		if (_dpiscale)
			y -= font.vdrawScaled(*this, x, y, color, *ch, _dpiscale);
		else
			y -= font.vdraw(*this, x, y, color, *ch);

	return y;
} // printf

/** print formatted string with shadow
 * @return right most pixel of the string
 */
int Painter::printfShadow(BFont& font, int x, int y, int ofs,
			const dword color, const dword shadow, const char *fmt, ...)
{
	char	text[1024];
	va_list	ap;
	if (fmt==NULL) return x;
	va_start(ap, fmt);
	vsprintf(text, fmt, ap);
	va_end(ap);

	for (byte *ch=(byte*)text; *ch; ch++) {
		if (_dpiscale) {
			     font.drawScaled(*this, x+ofs, y+ofs, shadow, *ch, _dpiscale);
			x += font.drawScaled(*this, x, y, color, *ch, _dpiscale);
		} else {
			     font.draw(*this, x+ofs, y+ofs, shadow, *ch);
			x += font.draw(*this, x, y, color, *ch);
		}
	}

	return x;
} // printfShadow

/** print formatted string with shadow
 * @return right most pixel of the string
 */
int Painter::printfOutline(BFont& font, int x, int y,
			const dword color, const dword outline, const char *fmt, ...)
{
	char	text[1024];
	va_list	ap;
	if (fmt==NULL) return x;
	va_start(ap, fmt);
	vsprintf(text, fmt, ap);
	va_end(ap);

	for (byte *ch=(byte*)text; *ch; ch++)
		x += font.drawOutline(*this, x, y, color, outline, *ch);

	return x;
} // printfShadow

/** Run a filter on all pixels having the flag FLAG_ALPHA */
void Painter::correctAntialias()
{
	dword* ptr = data();
	for (int j=0; j<_height; j++)
		for (int i=0; i<_width; i++, ptr++)
			if (*ptr & FLAG_ALPHA) {
				Color32 ncol;
				// find neighbors pixels if they are
				// normal colors and use them to blend it
				// 1st check next col (not changed yet)
				if (i<_width-1 and !(*(ptr+1)&FLAG_ALPHA))
					ncol = *(ptr+1);
				else	// 2nd one row down
				if (j<_height-1 and !(*(ptr+_width)&FLAG_ALPHA))
					ncol = *(ptr+_width);
				else	// 3rd previous col
				if (i>0 and !(*(ptr-1)&FLAG_ALPHA))
					ncol = *(ptr-1);
				else	// 4th previous row
				if (j>0 and !(*(ptr-_width)&FLAG_ALPHA))
					ncol = *(ptr-_width);
				else
					continue;

				Color32 col(*ptr);
				byte alpha = (col.alpha() & 0x7F) << 1;
				//int intensity = col.intensity();
				//col.lighter(intensity);
				col.blend(ncol, alpha);
				//col.alpha() = 0;
				*ptr = col & 0xFFFFFF;
			}
} // correctAntialias

/** run anti-aliasing + any other post-processing filter that might be wanted
* [1/17/2020 theis]
*/
void Painter::gaussianBlur()
{
	int dataPtr = 0;

	// NOTE: If one wants to increase speed a 5x5 Gaussian kernel could be used instead
	const float ppGaussKernel[7][7] = {
			{ 0.00000067f, 0.00002292f, 0.00019117f, 0.00038771f, 0.00019117f, 0.00002292f, 0.00000067f },
			{ 0.00002292f, 0.00078633f, 0.00655965f, 0.01330373f, 0.00655965f, 0.00078633f, 0.00002292f },
			{ 0.00019117f, 0.00655965f, 0.05472157f, 0.11098164f, 0.05472157f, 0.00655965f, 0.00019117f },
			{ 0.00038771f, 0.01330373f, 0.11098164f, 0.22508352f, 0.11098164f, 0.01330373f, 0.00038771f },
			{ 0.00019117f, 0.00655965f, 0.05472157f, 0.11098164f, 0.05472157f, 0.00655965f, 0.00019117f },
			{ 0.00002292f, 0.00078633f, 0.00655965f, 0.01330373f, 0.00655965f, 0.00078633f, 0.00002292f },
			{ 0.00000067f, 0.00002292f, 0.00019117f, 0.00038771f, 0.00019117f, 0.00002292f, 0.00000067f } };

/*
	const float ppGaussKernel[5][5] = {
			{ 0.003765f, 0.015019f, 0.023792f, 0.015019f, 0.003765f },
			{ 0.015019f, 0.059912f, 0.094907f, 0.059912f, 0.015019f },
			{ 0.023792f, 0.094907f, 0.150342f, 0.094907f, 0.023792f },
			{ 0.015019f, 0.059912f, 0.094907f, 0.059912f, 0.015019f },
			{ 0.003765f, 0.015019f, 0.023792f, 0.015019f, 0.003765f } };
*/

	// set up double buffer
	dword* pDblBuf = new dword[_maxSize];
	std::fill_n(pDblBuf, _maxSize, _background);

	// iterate over data & anti-alias where necessary
	// NOTE: we do not touch the original pixel to preserve it's color in full
	for (int x = 0; x < _width; x++) {
		for (int y = 0; y < _height; y++) {
			if ((pixel(x, y) & FLAG_ALPHA) == FLAG_ALPHA) {
				// for reasons of speed we currently ignore diagonal pixels,
				// though they could be included left & right
				// for edges we introduce a slight error here due to the wrap-around,
				// but it's hardly noticeable and avoids branching which might
				// slow down the process
				dataPtr = Clamp<int>(y * _width + x + 1, 0, _width * _height - 1);
				if ((_data[dataPtr] & FLAG_ALPHA) != FLAG_ALPHA)
					pDblBuf[dataPtr] = convolutePixel(x+1, y, ppGaussKernel);

				dataPtr = Clamp<int>(y * _width + x - 1, 0, _width * _height - 1);
				if ((_data[dataPtr] & FLAG_ALPHA) != FLAG_ALPHA)
					pDblBuf[dataPtr] = convolutePixel(x-1, y, ppGaussKernel);

				// up
				dataPtr = Clamp<int>((y+1) * _width + x, 0, _width * _height - 1);
				if ((_data[dataPtr] & FLAG_ALPHA) != FLAG_ALPHA)
					pDblBuf[dataPtr] = convolutePixel(x, y+1, ppGaussKernel);

				// down
				dataPtr = Clamp<int>((y-1) * _width + x, 0, _width * _height - 1);
				if ((_data[dataPtr] & FLAG_ALPHA) != FLAG_ALPHA)
					pDblBuf[dataPtr] = convolutePixel(x, y-1, ppGaussKernel);

				///////////////////////////////////////////////////////////////////
				// the following pixels could be discarded if one wants to improve
				// rendering speed up left & up right
				dataPtr = Clamp<int>((y+1) * _width + x - 1, 0, _width * _height - 1);
				if ((_data[dataPtr] & FLAG_ALPHA) != FLAG_ALPHA)
					pDblBuf[dataPtr] = convolutePixel(x-1, y+1, ppGaussKernel);

				dataPtr = Clamp<int>((y+1) * _width + x + 1, 0, _width * _height - 1);
				if ((_data[dataPtr] & FLAG_ALPHA) != FLAG_ALPHA)
					pDblBuf[dataPtr] = convolutePixel(x+1, y+1, ppGaussKernel);

				// down left & up right
				dataPtr = Clamp<int>((y-1) * _width + x - 1, 0, _width * _height - 1);
				if ((_data[dataPtr] & FLAG_ALPHA) != FLAG_ALPHA)
					pDblBuf[dataPtr] = convolutePixel(x-1, y-1, ppGaussKernel);

				dataPtr = Clamp<int>((y-1) * _width + x + 1, 0, _width * _height - 1);
				if ((_data[dataPtr] & FLAG_ALPHA) != FLAG_ALPHA)
					pDblBuf[dataPtr] = convolutePixel(x+1, y-1, ppGaussKernel);
			}
		}
	}

	// transfer double buffer to view buffer
	dword* pSSLine = pDblBuf;
	dword* pViewLine = data();
	for (int j = 0; j < _height; j++) {
		dword* pSSPtr = pSSLine;
		dword* pViewPtr = pViewLine;
		// TODO: check if a merging of the complete super-sampling buffer
		// and the image buffer is not faster than a conditional check
		for (int i = 0; i < _width; i++, pSSPtr++, pViewPtr++)
			if (*pSSPtr != background())
				*pViewPtr = *pSSPtr;

		pSSLine   += _width;
		pViewLine += _width;
	}

	delete [] pDblBuf;
} // gaussianBlur
