/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "matrix3.h"
#include <iomanip>

using std::endl;

constexpr double Matrix3::_identity[9];

bool Matrix3::isIdentity() const {
        return !memcmp(data, _identity, sizeof(data));
}

void Matrix3::identity() {
        memcpy(data, _identity, sizeof(data));
}

/** transpose */
void Matrix3::transpose()
{
	for (int i=0; i<3; i++)
		for (int j=i+1; j<3; j++) {
			double t = M(i,j);
			M(i, j) = M(j, i);
			M(j, i) = t;
		}
} // transpose

/** operator * *
Matrix3& Matrix3::operator *= (const Matrix3& mat)
{
	// FIXME Optimize
	return M = M*mat;
} * operator *= */

/** negate matrix */
void Matrix3::negate()
{
	for (int i=0; i<3; i++)
		for (int j=0; j<3; j++)
			M(i,j) = -M(i,j);
} // negate

/** invert the matrix in place and
 * @return true on success, false otherwise
 */
bool Matrix3::inverse(const double eps)
{
	Matrix3 inv;
	double D = det();
	if (Eq0(D, eps)) return false;
	D = 1.0/D;

	inv(0,0) =  (M(1,1)*M(2,2) - M(2,1)*M(1,2)) * D;
	inv(0,1) = -(M(0,1)*M(2,2) - M(0,2)*M(2,1)) * D;
	inv(0,2) =  (M(0,1)*M(1,2) - M(0,2)*M(1,1)) * D;
	inv(1,0) = -(M(1,0)*M(2,2) - M(1,2)*M(2,0)) * D;
	inv(1,1) =  (M(0,0)*M(2,2) - M(0,2)*M(2,0)) * D;
	inv(1,2) = -(M(0,0)*M(1,2) - M(1,0)*M(0,2)) * D;
	inv(2,0) =  (M(1,0)*M(2,1) - M(2,0)*M(1,1)) * D;
	inv(2,1) = -(M(0,0)*M(2,1) - M(2,0)*M(0,1)) * D;
	inv(2,2) =  (M(0,0)*M(1,1) - M(1,0)*M(0,1)) * D;
	copy(inv);
	return true;
} // inverse

/** rotate matrix around an axis
 * @param angle
 * @param axis 0=X, 1=Y, 2=Z
 * */
void Matrix3::rotate(const double angle, const int axis)
{
	double s, c;
	identity();

	bsincos(angle, &s, &c);

	int m1 = ((axis+1)%3)+1;
	int m2 = m1%3;
	m1 -= 1;

	M(m1,m1) =  c;
	M(m2,m2) =  c;
	M(m1,m2) = -s;
	M(m2,m1) =  s;
} // rotate

/** rotate matrix around a vector */
void Matrix3::rotate(const double angle, double x, double y, double z)
{
	double s, c;
	identity();

	bsincos(angle, &s, &c);

	double invl = 1.0/sqrt(x*x + y*y + z*z);
	x *= invl;
	y *= invl;
	z *= invl;

	double c1 = 1.0 - c;
	M(0,0) = x*x + (1-x*x)*c;
	M(0,1) = x*y*c1 - z*s;
	M(0,2) = x*z*c1 + y*s;

	M(1,0) = x*y*c1 + z*s;
	M(1,1) = y*y + (1-y*y)*c;
	M(1,2) = y*z*c1 - x*s;

	M(2,0) = x*z*c1 - y*s;
	M(2,1) = y*z*c1 + x*s;
	M(2,2) = z*z + (1-z*z)*c;
} // rotate

/** @return rank  of matrix
 * Warning: destroys the matrix as it applies a rref (reduced row echelon form*/
int Matrix3::rank(const double eps)
{
	bool row_selected[3] = {false, false, false};

	int rank_ = 0;
	for (int i=0; i<cols(); ++i) {
		int j;
		for (j=0; j<rows(); ++j)
			if (!row_selected[j] && !Eq0(M(j,i),eps))
				break;

		if (j != rows()) {
			++rank_;
			row_selected[j] = true;
			for (int p=i+1; p<cols(); ++p)
				M(j,p) /= M(j,i);
			for (int k=0; k<rows(); ++k)
				if (k!=j && !Eq0(M(k,i),eps))
					for (int p=i+1; p<cols(); ++p)
						M(k,p) -= M(j,p) * M(k,i);
		}
	}
	return rank_;
} // rank

/** @return Eigenvalues of a symmetric matrix
 * Given a real symmetric 3x3 matrix A, compute the eigenvalues
 * Note that acos and cos operate on angles in radians
 * @see https://en.wikipedia.org/wiki/Eigenvalue_algorithm
 */
void Matrix3::eigenValuesOfSymmetric(double eig[]) const
{
	double p1 = Sq(M(0,1)) + Sq(M(0,2)) + Sq(M(1,2));
	if (Eq0(p1,Vector::_epsilon)) {			// M is diagonal.
		eig[0] = M(0,0);
		eig[1] = M(1,1);
		eig[2] = M(2,2);
	} else {
		double q  = trace()/3;		// trace(M) is the sum of all diagonal values
		double p2 = Sq(M(0,0)-q) + Sq(M(1,1)-q) + Sq(M(2,2)-q) + 2*p1;
		double p  = sqrt(p2/6);
		Matrix3 I(1);
		Matrix3 B = (1/p) * (*this - q*I);	// I is the identity matrix
		double r  = B.det()/2;

		// In exact arithmetic for a symmetric matrix  -1 <= r <= 1
		// but computation error can leave it slightly outside this range.
		double phi;
		if (r <= -1)
			phi = PI / 3;
		else
		if (r >= 1)
			phi = 0;
		else
			phi = acos(r) / 3;

		// the eigenvalues satisfy eig[2] <= eig[1] <= eig[0]
		eig[0] = q + 2*p*cos(phi);
		eig[2] = q + 2*p*cos(phi + (2*PI/3));
		eig[1] = 3*q - eig[0] - eig[2];	// since trace(M) = eig[0] + eig[1] + eig[2]
	}
} // eigenValuesOfSymmetric

/** matrix print out */
std::ostream& operator << (std::ostream& s, const Matrix3& matrix)
{
	for (int r=0; r<matrix.rows(); r++) {
		if (r==0)
			s << "/";
		else
		if (r+1==matrix.rows())
			s << "\\";
		else
			s << "|";
		for (int c=0; c<matrix.columns(); c++) {
			s << " " << std::setw(16) << std::setprecision(10) << matrix(r,c);
		}
		if (r==0)
			s << " \\" << endl;
		else
		if (r+1==matrix.rows())
			s << " /" << endl;
		else
			s << " |" << endl;
	}
	return s;
} /* operator << */
