/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
// Authors:	Christian.Theis@cern.ch
//		Vasilis.Vlachoudis@cern.ch
 * Date:	14-Jul-2021
 */

#ifndef __MATH_MORTON_H
#define __MATH_MORTON_H

// we add a 0 on position 1 in this array to use same code for 32-bit and 64-bit cases
constexpr unsigned	cMortonMask32[6] = {
					0x000003ff,
					0,
					0x30000ff,
					0x0300f00f,
					0x30c30c3,
					0x9249249 };
constexpr unsigned long long  cMortonMask64[6] = {
					0x1fffff,
					0x1f00000000ffff,
					0x1f0000ff0000ff,
					0x100f00f00f00f00f,
					0x10c30c30c30c30c3,
					0x1249249249249249 };

/* ============================== Morton ============================== */
/** Simple Morton encoder to map 2D coordinates to a space filling Z-curve */
class Morton {
public:
	static bool	next2D(int& x, int& y, const int width, const int height);

	static unsigned encode2D(unsigned x, unsigned y)
			{ return (interleave(y) << 1) + interleave(x); };

#if _DEBUG>0
	static void	test(int W, int H, bool plot);
#endif
protected:
	static unsigned interleave(unsigned x) {
			x &= 0x0000ffff;
			x = (x ^ (x << 8)) & 0x00ff00ff;
			x = (x ^ (x << 4)) & 0x0f0f0f0f;
			x = (x ^ (x << 2)) & 0x33333333;
			x = (x ^ (x << 1)) & 0x55555555;
			return x;
		}
}; // Morton

/* ============================== Morton =============================== */
/** Simple Morton encoder of 3D coordinates working with float or doubles*/
// https://www.forceflow.be/2013/10/07/morton-encodingdecoding-through-bit-interleaving-implementations/
//
class Morton3 {
public:
	template<typename TMorton, typename TCoordinate>
	static TMorton encode3D( const TCoordinate x, const TCoordinate y, const TCoordinate z ) {
		return interleave3<TMorton, TCoordinate>(x)
			| (interleave3<TMorton, TCoordinate>(y) << 1)
			| (interleave3<TMorton, TCoordinate>(z) << 2);
	}

protected:
	template<typename TMorton, typename TCoordinate>

	static	TMorton interleave3( const TCoordinate a ) {
			const TMorton* pMasks = (sizeof( TMorton ) <= 4) ?
					reinterpret_cast<const TMorton*>(cMortonMask32) :
					reinterpret_cast<const TMorton*>(cMortonMask64);
			TMorton x = ((TMorton)a) & pMasks[0];

			// NOTE: this is a change from a std. morton encoding and
			// required to handle border cases of our linear BVH where we
			// might have centroids of zones at an relative offset of 1.0
			// wrt to the overall bbox. in a 10 bit encoding we would have
			// an overrun which we have to handle
			if (x == (1<<10)) --x;

			if (sizeof(TMorton)==8)
				x = (x | (unsigned long long)x << 32) & pMasks[1];	// for 64-bit case
			x = (x | x << 16) & pMasks[2];
			x = (x | x <<  8) & pMasks[3];
			x = (x | x <<  4) & pMasks[4];
			x = (x | x <<  2) & pMasks[5];
			return x;
		}
}; // class Morton3

#endif
