/**
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	10-Oct-2018
 */

#ifndef __TOOL_CACHE_H
#define __TOOL_CACHE_H

#include <vector>

// =================================== Cache ===================================
/**
 * A template class to cache calculations of items. Items must inherit
 * from @see CacheItem.
 * Before a new calculation call the inc() operator++ to increase the
 * checkid. To check call the isOk() methods to verify the integrity
 * of the CacheItem
 */
template <typename T>
class Cache final : public std::vector<T> {
private:
	int	gid;		// current checkid
	int	gmaxid;		// maximum checkid for all items

public:
	Cache() : gid(0), gmaxid(0) {}
	Cache(const size_t n) : gid(0), gmaxid(0) { resize(n); }
	~Cache() {}

	// safety catches allowing for exception handling when empty cache structures are used
#if _DEBUG>0
	T& operator [](int i)		  { return std::vector<T>::at(i); }	// range-checked
	const T& operator [](int i) const { return std::vector<T>::at(i); }	// range-checked
#endif

	/** resize array to the new number of bodies */
	void	resize(const size_t n) {
				std::vector<T>::resize(n, T());
				for (size_t i=n; i--; )
					(*this)[i].reset();
			}

	/** Increase body check id. Needed when we want to calculate for the bodies
	 *  a new location
	 */
	int	inc() {
			int old = gid;
			gid = ++gmaxid;
			return old;
		}
	int	operator++()		{ return inc(); }
	int	operator()() const	{ return gid; }
//	void	set(int value)		{ gid = value; }

	void	mark(int i)		{ assert(i < this->size()); (*this)[i].set(gid); }
	void	mark(T& t)		{ t.set(gid); }
	bool	isOk(int i) const	{ assert(i < this->size()); return (*this)[i]() == gid; }
	bool	isOk(T& t) const	{ return t() == gid; }
}; // class Cache

#if 0
// =================================== Cache ===================================
template <typename T>
class Cache2 /*: public std::vector<std::pair<int,T>> */ {
private:
	int	gid;				// current checkid
	int	gmaxid;				// maximum checkid for all items
	std::vector<std::pair<int,T> >	items;	// array of items to be cached

public:
	Cache2() : gid(0), gmaxid(0) {}
	~Cache2() {}

	/** resize array to the new number of bodies */
	void	resize(const size_t n) {
				items.resize(n, std::make_pair<int,T>(-1,T()));
				for (size_t i=0; i<n; i++)	// ??
					items[i].first = -1;
			}

	int	operator()() const	{ return gid; }
//	void	set(int value)		{ gid = value; }

//	T&	operator[](int i)	{ return items[i].second; }
//	std::pair<int,T>& operator[](int i)	{ return items[i]; }
	T&	operator[](int i)	{ return items[i].second; }
	int	first(int i)		{ return items[i].first; }

	/** Increase body check id. Needed when we want to calculate for the bodies
	 *  a new location
	 */
	int	inc() {
			int old = gid;
			gid = ++gmaxid;
			return old;
		}
	int	operator++()		{ return inc(); }

	void	mark(int i)		{ items[i].first = gid; }
//	void	mark(T& t)		{ t(gid); }
	bool	ischecked(int i) const	{ return items[i].first == gid; }
//	bool	ischecked(T& t) const	{ return t() == gid; }
}; // class Cache2
#endif

// ================================= CacheItem =================================
class CacheItem {
private:
	int	_id;
public:
	CacheItem() : _id(-1) {}
	~CacheItem() {}

	/** reset check id */
	void	reset()				{ _id  = -1; }

	/** @return check id */
	int	operator()()		const	{ return _id; }
	int	get()			const	{ return _id; }

	/** set _id */
	void	operator()(int gid)		{ _id = gid; }
	void	set(int gid)			{ _id = gid; }

//	/** mark as checked using cache */
//	void	mark(Cache<_CacheItem>& cache)		{ _id = cache(); }
//	/** @return true if it is checked already */
//	bool	isChecked(Cache<_CacheItem>& cache) const	{ return _id == cache(); }
}; // class CacheItem

#endif
