/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#include "imagelayer.h"
#include "viewer.h"

/** ImageLayer */
ImageLayer::ImageLayer(const Geometry& g, GeometryKernel& k, GeometryViewer& v) : Layer(g, k, v)
{
	// background image
	_data    = NULL;
	_size    = 0;
	_width   = 0;
	_height  = 0;
	_alpha   = 0x7F;
	_minW    = -INFINITE;
	_maxW    = INFINITE;
	_visible = false;
	R.identity();
	M.identity();
} // ImageLayer

/** ~ImageLayer */
ImageLayer::~ImageLayer()
{
	if (_size>0) delete [] _data;
} /* ~ImageLayer */

/** data */
bool ImageLayer::data(dword* dat, size_t siz)
{
	if (siz==0) {
		// set the pointer DO NOT COPY
		if (_size>0) delete [] _data;
		_data = dat;
		_size = 0;
		return true;
	} else
	if (siz != _size) {
		if (_data) delete [] _data;
		try {
			_data = new dword[siz];
			_size = siz;
		} catch (...) {
			_data = NULL;
			_size = 0;
			return false;
		}
	}
	memcpy(_data, dat, siz);
	return true;
} // data

/** matrix
 * set image Matrix and prepare the inverse matrix and check for visibility
 */
void ImageLayer::matrix(const Matrix3* r, const Matrix3* m)
{
	if (r != NULL) R.copy(*r);
	if (m != NULL) M.copy(*m);

	// Calculate inverse and image matrices
	//_viewMatrix.multiply(_matrix, view().matrix());

	/* Check matrix and view matrix are on the same projection */
	Vector imgW(view().invMatrix(2,0), view().invMatrix(2,1), view().invMatrix(2,2));
	Vector w(M(2,0), M(2,1), M(2,2));
	_visible = Abs(imgW*w)/w.length() > 0.9999;
} // matrix

/** colorRange */
void ImageLayer::colorRange(dword low, dword high)
{
	Color32 lw(low), hg(high);
	Color32* ptr = (Color32*)_data;
	for (size_t i=0; i<size(); i++, ptr++)
		ptr->remap(lw, hg);
} // colorRange

/** findBackground */
void ImageLayer::findBackground()
{
	int sum[3*256];
	dword color[3*256];
	memset(sum, 0, sizeof(sum));
	memset(color, 0, sizeof(color));
	dword* d = _data;
	for (size_t i=0; i<size(); i++) {
		int s = (*d & 0xff) + (*d>>8 & 0xff) + (*d>>16 & 0xff);
		sum[s]++;
		color[s] = *d;
		d++;
	}
	int mi = 0;
	for (int i=0; i<3*256; i++)
		if (sum[i]>sum[mi])
			mi = i;

	_background = color[mi];
} // findBackground

/** draw
 * FIXME 1. Can be optimized by introducing a bounding box around the image and calculate
 *          first the intersection of the 6 planes
 *       2. Modified to check for non-lines, in order to draw also the regions
 *       3. Rotation of the USRBIN
 */
void ImageLayer::draw(Painter& painter)
{
	if (!_visible) return;

	Color32* ptr = (Color32*)painter.data();
	int H = painter.height();	// save variables that might change
	int W = painter.width();	// during the scanning...

	for (int j=0; j<H; j++) {
		if (stop()) break;
		double v = view().jc2v(j);
		for (int i=0; i<W; i++, ptr++) {
			dword flag = *ptr & FLAG_INFOMASK;
			if (flag==0 || flag&(FLAG_FINAL|FLAG_ALPHA)) continue;

			double u = view().ic2u(i);
			Point  p = view().uv2xyz(u,v);

			double ui = M(0,0)*p.x + M(0,1)*p.y + M(0,2)*p.z;
			double vi = M(1,0)*p.x + M(1,1)*p.y + M(1,2)*p.z;
			double xi = R(0,0)*ui + R(0,1)*vi + R(0,2);
			double yi = R(1,0)*ui + R(1,1)*vi + R(1,2);

			int pi = (int)xi;
			int pj = (int)yi;

			if (pi<0 || pi>=_width || pj<0 || pj>=_height) continue;
			int loc = pj*_width + pi;

			Color32 pixel, pixelR, pixelD, pixelDR;
			pixel = _data[loc];
			if (pi < _width-1)
				pixelR = _data[loc+1];
			else
				pixelR = pixel;

			if (pj < _height-1) {
				pixelD = _data[loc+_width];
				if (pi < _width-1)
					pixelDR = _data[loc+_width+1];
				else
					pixelDR = pixelR;
			} else {
				pixelD  = pixel;
				pixelDR = pixelR;
			}

			double a = xi - (double)pi;
			double b = yi - (double)pj;
			double a1 = 1.0 - a;
			double b1 = 1.0 - b;

			// Bilinear smoothing
			int red   = (int)(a1*b1*(double)pixel.red() +
			                  a *b1*(double)pixelR.red() +
			                  a1*b *(double)pixelD.red() +
			                  a *b *(double)pixelDR.red());
			int green = (int)(a1*b1*(double)pixel.green() +
			                  a *b1*(double)pixelR.green() +
			                  a1*b *(double)pixelD.green() +
			                  a *b *(double)pixelDR.green());
			int blue  = (int)(a1*b1*(double)pixel.blue() +
			                  a *b1*(double)pixelR.blue() +
			                  a1*b *(double)pixelD.blue() +
			                  a *b *(double)pixelDR.blue());
			if (pixel.alpha() > 0) {
				ptr->alpha() = _alpha;
				ptr->blend(Color32(red,green,blue,0xFF));
				*ptr |= FLAG_REGION;
			}
		}
	}
} // draw
