/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	01-Apr-2011
 */

#include "userdumplayer.h"

#include <sys/stat.h>

#include "viewer.h"

UserDumpLayer::UserDumpLayer(const Geometry& g, GeometryKernel& k, GeometryViewer& v)
	: Layer(g, k, v),
	  _mtime(0),
	  alpha(0)
{
	reset();
} // UserDumpLayer

UserDumpLayer::~UserDumpLayer()
{
	cleanup();
} // ~UserDumpLayer

/** cleanup */
void UserDumpLayer::cleanup()
{
	_filename.clear();
	_mtime = 0;
} // cleanup

/** reset */
void UserDumpLayer::reset()
{
	Random random(314159265L);
	for (int i=0; i<NPARTICLES; i++) {
		_color[i] = random.integer();
		_emin[i]  = 0.0;
		_emax[i]  = 1e99;
	}
} // reset

/** open filename for userdump plotting */
bool UserDumpLayer::open(const char* name)
{
	struct stat st;

	if (_filename.compare(name))
		cleanup();

	// Check if information is valid
	if (stat(name, &st)) {
		cleanup();
		return true;
	}
	if (st.st_mtime==_mtime) return false;

	cleanup();
	_filename = name;
	if (userdump.open(name)) return false;

	// Remember modification time
	_mtime = st.st_mtime;

	return true;
} // filename

/** draw */
void UserDumpLayer::draw(ZPainter& painter, bool show3D)
{
	if (!userdump) return;
	userdump.seek(0);

	int i = 0;
	while (i<start && userdump.readEvent(0)) {
		if (stop()) return;
		if (userdump.type == UserDump::MGDRAW_SOURCE) i++;
	}

	i = n+1;
	while (i>0 && userdump.readEvent(-1)) {
		if (stop()) return;
		if (userdump.type == UserDump::MGDRAW_TRACK) {
			int particle = userdump.event.track.particle;
			dword col = color(particle);
			if ((col&COLOR_ALPHA) == COLOR_ALPHA) continue;
			if (!InRange(	emin(particle),
					(double)userdump.event.track.energy,
					emax(particle))) continue;
			UserDump::TrackPos& prev = userdump.track[0];
			Point A;
			double up=0, vp=0;	// set to zero to keep compiler happy
			if (show3D)
				A.set(prev.x, prev.y, prev.z);
			else {
				up = view().xyz2u(prev.x, prev.y, prev.z);
				vp = view().xyz2v(prev.x, prev.y, prev.z);
			}
			for (size_t j=1; j<userdump.track.size(); j++) {
				UserDump::TrackPos& pos = userdump.track[j];
				if (show3D) {
					Point B(pos.x, pos.y, pos.z);
					draw3Dline(painter,A,B,col,alpha);
					A = B;
				} else {
					double u = view().xyz2u(pos.x, pos.y, pos.z);
					double v = view().xyz2v(pos.x, pos.y, pos.z);
					double cup = up;	// clip coordinates
					double cvp = vp;
					double cu  = u;
					double cv  = v;
					if (view().clipLine(&cup, &cvp, &cu, &cv)) {
						painter.line(view().u2i(cup), view().v2j(cvp),
							     view().u2i(cu),  view().v2j(cv),
							     col);
					}
					up = u;
					vp = v;
				}
			}
		} else
		if (userdump.type == UserDump::MGDRAW_SOURCE)
			i--;
	}
} // draw
